from __future__ import annotations

from email.mime.multipart import MIMEMultipart
from email.contentmanager import raw_data_manager
from email.message import EmailMessage
from typing import TYPE_CHECKING

from e3.aws.cfn import Resource, AWSType, GetAtt, Base64, Join, Ref, Sub
from e3.aws.cfn.iam import PolicyDocument
from e3.aws.ec2.ami import AMI

if TYPE_CHECKING:
    from e3.aws.cfn.ec2.security import SecurityGroup

CFN_INIT_STARTUP_SCRIPT = """#!/bin/sh
sed -i 's/scripts-user$/[scripts-user, always]/' /etc/cloud/cloud.cfg
${Cfninit} -v --stack ${AWS::StackName} \\
                --region ${AWS::Region} \\
                --resource ${Resource} \\
                --configsets ${Config} ${CfninitOptions}\n\n"""


CFN_INIT_STARTUP_SCRIPT_WIN = (
    "C:\\ProgramData\\Amazon\\EC2-Windows\\"
    + "Launch\\Scripts\\InitializeInstance.ps1 -schedule \n"
    + "${Cfninit} -v --stack ${AWS::StackName} --region "
    + "${AWS::Region} --resource ${Resource} --configsets ${Config} "
    + "${CfninitOptions}\n\n"
)


class BlockDevice:
    """Block device for EC2 instances."""

    pass


class EphemeralDisk(BlockDevice):
    """Ephemeral disk."""

    def __init__(self, device_name, id=0):
        """Initialize an ephemeral disk.

        :param device_name: name of the device associated with that disk
        :type device_name: str
        :param id: id of the ephemeral disk (default is 0)
        :type id: int
        """
        assert isinstance(id, int)
        self.device_name = device_name
        self.id = id

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        return {"DeviceName": self.device_name, "VirtualName": "ephemeral%s" % self.id}


class EBSDisk(BlockDevice):
    """EBS Disk."""

    def __init__(self, device_name, size=None, encrypted=None):
        """Initialize an EBS disk.

        :param device_name: name of the device associated with that disk
        :type device_name: str
        :param size: disk size in Go (default: 20Go). None can be used to
            use the same size as the original AMI
        :type size: int | None
        :param encrypted: if True encrypt the device, if None take the default
            (useful when device is created from a snapshot).
        :type encrypted: bool | None
        """
        self.device_name = device_name
        self.size = size
        self.encrypted = encrypted

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        result = {
            "DeviceName": self.device_name,
            "Ebs": {"VolumeType": "gp2", "DeleteOnTermination": True},
        }
        if self.size is not None:
            result["Ebs"]["VolumeSize"] = str(self.size)

        if self.encrypted is not None:
            result["Ebs"]["Encrypted"] = self.encrypted
        return result


class EC2NetworkInterface:
    """EC2 Instance network interface."""

    def __init__(
        self,
        subnet=None,
        public_ip=False,
        groups=None,
        device_index=None,
        description=None,
        interface=None,
    ):
        """Initialize a EC2NetworkInterface.

        :param subnet: subnet to which the interface is attached
        :type subnet: e3.aws.cfn.ec2.Subnet
        :param public_ip: if True assign automatically public IP address.
            Default is False.
        :type public_ip: bool
        :param groups: list of security groups associated with the interface.
            If no group is specified, AWS will assign a default group.
        :type groups: list[SecurityGroup] | None
        :param device_index: natural giving the interface position. 0 is the
            default interface. If set to None, some method such as
            e3.aws.cfn.ec2.Instance.add will assign automatically a device
            index
        :type device_index: 0 | None
        :param description: optional description
        :type description: str | None
        :param interface: an external network interface. If specified subnet,
            public_ip and groups should be set to None
        :type interface: NetworkInterface | None
        """
        if subnet is not None:
            assert isinstance(subnet, Subnet), "unexpected type for subnet: %s" % subnet
            assert (
                interface is None
            ), "cannot specify a network interface if subnet is set"
            self.subnet = subnet
            self.public_ip = public_ip
            self.groups = groups
            self.interface_id = None
        else:
            assert isinstance(interface, NetworkInterface), (
                "if not subnet is provided a valid network interface "
                "should be passed"
            )
            assert not public_ip, "cannot associate automatically a public IP"
            assert groups is None, "groups should be set in the network interface"
            self.interface = interface
            self.subnet = None
            self.public_ip = False
            self.groups = None

        self.device_index = device_index
        self.description = description

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        result = {}

        if self.subnet:
            result["AssociatePublicIpAddress"] = self.public_ip
            result["SubnetId"] = self.subnet.ref
            result["DeleteOnTermination"] = True
        else:
            result["NetworkInterfaceId"] = self.interface.ref

        if self.device_index is not None:
            result["DeviceIndex"] = self.device_index
        if self.groups:
            result["GroupSet"] = [group.ref for group in self.groups]
        if self.description is not None:
            result["Description"] = self.description
        return result


class UserData:
    """EC2 Instance user data."""

    def __init__(self):
        """Initialize user data."""
        self.parts = []
        self.variables = {}

    def add(self, kind, content, name, variables=None):
        """Add an entry in the user data.

        :param kind: MIME subtype (maintype is always text)
        :type kind: str
        :param content: the content associated with that value
        :type content: str
        :param name: name of the entry (aka filename)
        :type name: str
        """
        if variables is not None:
            self.variables.update(variables)
        self.parts.append((name, kind, content))

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        # This is important to keep the boundary static in order to avoid
        # spurious instance reboots.
        multi_part = MIMEMultipart(
            boundary="-_- :( :( /o/ Static User Data Boundary /o/ :) :) -_-"
        )
        for name, kind, part in self.parts:
            mime_part = EmailMessage()
            raw_data_manager.set_content(mime_part, part, subtype=kind, filename=name)
            multi_part.attach(mime_part)
        return Base64(Sub(multi_part.as_string(), self.variables))


class WinUserData:
    """EC2 Windows Instance user data."""

    def __init__(self):
        """Initialize user data."""
        self.parts = []
        self.variables = {}

    def add(self, kind, content, variables=None):
        """Add an entry in the user data.

        :param kind: script/powershell/persist
        :type kind: str
        :param content: the content associated with that value
        :type content: str
        """
        if variables is not None:
            self.variables.update(variables)
        self.parts.append((kind, content))

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        props = ""
        for kind, part in self.parts:
            props += "<%s>\n%s\n</%s>" % (kind, part, kind)
        return Base64(Sub(props, self.variables))


class NetworkInterface(Resource):
    """External Network Interface."""

    def __init__(self, name, subnet, groups=None, description=None):
        """Initialize an External Network Interface (ENI).

        :param name: logical name of the instance
        :type name: str
        :param subnet: subnet to which the interface is attached
        :type subnet: e3.aws.cfn.ec2.Subnet
        :param groups: list of security groups associated with the interface.
            If no group is specified, AWS will assign a default group.
        :type groups: list[SecurityGroup] | None
        :param description: optional description
        :type description: str | None
        """
        super().__init__(name, kind=AWSType.EC2_NETWORK_INTERFACE)
        assert isinstance(subnet, Subnet)
        self.subnet = subnet
        self.groups = groups
        self.description = description
        self.tags = {}

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        result = {"SubnetId": self.subnet.ref}
        if self.description is not None:
            result["Description"] = self.description
        if self.groups is not None:
            result["GroupSet"] = [group.ref for group in self.groups]
        if self.tags:
            result["Tags"] = [{"Key": k, "Value": v} for k, v in self.tags.items()]
        return result


class TemplateOrInstance(Resource):
    def set_instance_profile(self, profile):
        self.instance_profile = profile

    def add(self, device):
        """Add a device to the instance.

        :param device: can be a disk or a network interface
        :type device: EC2NetworkInterface | BockDevice
        :return: the Instance itself
        :rtype: Instance
        """
        if isinstance(device, EC2NetworkInterface):
            if device.device_index is None:
                # Assign automatically a device index
                index = max(list(self.network_interfaces.keys()) + [-1]) + 1
                device.device_index = index
            else:
                # Ensure the device is not already present
                assert device.device_index not in self.network_interfaces
                index = device.device_index

            self.network_interfaces[index] = device
        elif isinstance(device, BlockDevice):
            self.block_devices.append(device)
        else:
            raise AssertionError("invalid device %s" % device)
        return self

    def add_user_data(self, kind, content, name=None, variables=None):
        """Add a user data entry.

        :param kind: MIME subtype (maintype is always text)
        :type kind: str
        :param content: the content associated with that value
        :type content: str
        :param name: name of the entry (aka filename)
        :type name: str
        """
        if self.image.is_windows:
            assert name is None
            if self.user_data is None:
                self.user_data = WinUserData()
            self.user_data.add(kind, content, variables=variables)
        else:
            assert name is not None
            if self.user_data is None:
                self.user_data = UserData()
            self.user_data.add(kind, content, name, variables=variables)

    def set_cfn_init(
        self,
        config="init",
        cfn_init="/usr/local/bin/cfn-init",
        resource=None,
        metadata=None,
        init_script="",
        use_instance_role=False,
    ):
        """Add CFN init call on first boot of the instance.

        :param config: name of the configset to be launch (default: init)
        :type config: str
        :param cfn_init: location of cfn-init on the instance
            (default: /usr/local/bin/cfn-init)
        :type cfn_init: str
        :param resource: resource in which the metadata will be added. Default
            is to use current resource
        :type resource: str | None
        :param metadata: dict conforming to AWS::CloudFormation::Init
            specifications
        :type metadata: dict | None
        :param init_script: command to launch after cfn-init
        :type init_script: powershell command for windows and bash command for
            linuxes
        """
        if resource is None:
            resource = self.name

        if use_instance_role:
            cfn_init_options = Join([" --role ", self.instance_profile.role])
        else:
            cfn_init_options = ""

        if self.image.is_windows:
            self.add_user_data(
                "powershell",
                CFN_INIT_STARTUP_SCRIPT_WIN + init_script,
                variables={
                    "Cfninit": cfn_init,
                    "Config": config,
                    "Resource": resource,
                    "CfninitOptions": cfn_init_options,
                },
            )
            self.add_user_data("persist", "true")
        else:
            self.add_user_data(
                "x-shellscript",
                CFN_INIT_STARTUP_SCRIPT + init_script,
                "init.sh",
                variables={
                    "Cfninit": cfn_init,
                    "Config": config,
                    "Resource": resource,
                    "CfninitOptions": cfn_init_options,
                },
            )

        if metadata is not None:
            self.metadata["AWS::CloudFormation::Init"] = metadata


class LaunchTemplate(TemplateOrInstance):
    """EC2 Launch template."""

    def __init__(
        self,
        name,
        image,
        instance_type="t2.micro",
        disk_size=None,
        terminate_on_shutdown=False,
        template_name=None,
        copy_ami_tags=True,
    ):
        """Initialize an EC2 launch template.

        :param name: logical name of the instance
        :type name: str
        :param image: AMI to use
        :type image_id: e3.aws.ec2.ami.AMI
        :param instance_type: kind of instance (default t2.micro)
        :type instance_type: str
        :param disk_size: size of disk. If None the disk size will be
            the original AMI one. Note that this affect only the root
            device of the AMI
        :type disk_size: int | None
        :param terminate_on_shutdown: if True the instance is terminated on
            shutdown
        :type terminate_on_shutdown: bool
        :param template_name: if not None set the template name. If None
            logical resource id will be used for the template name
        :type template_name: str | None
        :param copy_ami_tags: if True AMI tags will be copied into the template
            tags
        :type copy_ami_tags: bool
        """
        super().__init__(name, kind=AWSType.EC2_LAUNCH_TEMPLATE)
        assert isinstance(image, AMI)
        self.image = image
        self.instance_type = instance_type
        self.block_devices = []
        if disk_size is not None:
            self.add(EBSDisk(device_name=self.image.root_device, size=disk_size))
        self.instance_profile = None
        self.network_interfaces = {}
        self.user_data = None
        self.tags = {}
        self.terminate_on_shutdown = terminate_on_shutdown
        if template_name is None:
            self.template_name = self.name
        else:
            self.template_name = template_name
        self.copy_ami_tags = copy_ami_tags

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        td = {
            "ImageId": self.image.id,
            "InstanceType": self.instance_type,
            "BlockDeviceMappings": [bd.properties for bd in self.block_devices],
        }

        if self.instance_profile is not None:
            td["IamInstanceProfile"] = {"Name": self.instance_profile.ref}

        if self.network_interfaces:
            td["NetworkInterfaces"] = []
            for ni in self.network_interfaces.values():
                # We need to tweak a bit the result for the network
                # interface as API is not coherent between Instances and
                # Templates. Basically the key GroupSet should be replaced
                # by Groups
                prop = ni.properties
                if "GroupSet" in prop:
                    prop["Groups"] = prop["GroupSet"]
                    del prop["GroupSet"]
                td["NetworkInterfaces"].append(prop)

        if self.user_data is not None:
            td["UserData"] = self.user_data.properties

        if self.tags:
            merged_tags = {}
            if self.copy_ami_tags:
                merged_tags.update(self.image.tags)
            merged_tags.update(self.tags)

            tags = [{"Key": k, "Value": v} for k, v in merged_tags.items()]

            # Tag both volumes and instance
            td["TagSpecifications"] = [
                {"ResourceType": "instance", "Tags": tags},
                {"ResourceType": "volume", "Tags": tags},
            ]

        if self.terminate_on_shutdown:
            td["InstanceInitiatedShutdownBehavior"] = "terminate"

        return {"LaunchTemplateData": td, "LaunchTemplateName": self.template_name}


class Instance(TemplateOrInstance):
    """EC2 Instance."""

    ATTRIBUTES = (
        "AvailabilityZone",
        "PrivateDnsName",
        "PublicDnsName",
        "PrivateIp",
        "PublicIp",
    )

    def __init__(
        self,
        name,
        image,
        instance_type="t2.micro",
        disk_size=None,
        terminate_on_shutdown=False,
        copy_ami_tags=False,
    ):
        """Initialize an EC2 instance.

        :param name: logical name of the instance
        :type name: str
        :param image: AMI to use
        :type image_id: e3.aws.ec2.ami.AMI
        :param instance_type: kind of instance (default t2.micro)
        :type instance_type: str
        :param disk_size: size of disk. If None the disk size will be
            the original AMI one. Note that this affect only the root
            device of the AMI
        :type disk_size: int | None
        :param terminate_on_shutdown: if True the instance is terminated on
            shutdown
        :type terminate_on_shutdown: bool
        :param copy_ami_tags: if True AMI tags will be copied into the template
            tags
        :type copy_ami_tags: bool
        """
        super().__init__(name, kind=AWSType.EC2_INSTANCE)
        assert isinstance(image, AMI)
        self.image = image
        self.instance_type = instance_type
        self.block_devices = []
        if disk_size is not None:
            self.add(EBSDisk(device_name=self.image.root_device, size=disk_size))
        self.instance_profile = None
        self.network_interfaces = {}
        self.user_data = None
        self.tags = {}
        self.copy_ami_tags = copy_ami_tags
        self.terminate_on_shutdown = terminate_on_shutdown

    @property
    def public_ip(self):
        """Return a reference to the public Ip.

        :rtype: e3.aws.cfn.GetAtt
        """
        return GetAtt(self.name, "PublicIp")

    @property
    def private_ip(self):
        """Return a reference to the private Ip.

        :rtype: e3.aws.cfn.GetAtt
        """
        return GetAtt(self.name, "PrivateIp")

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        result = {
            "ImageId": self.image.id,
            "InstanceType": self.instance_type,
            "BlockDeviceMappings": [bd.properties for bd in self.block_devices],
        }

        if self.instance_profile is not None:
            result["IamInstanceProfile"] = self.instance_profile.ref

        if self.network_interfaces:
            result["NetworkInterfaces"] = [
                ni.properties for ni in self.network_interfaces.values()
            ]

        if self.user_data is not None:
            result["UserData"] = self.user_data.properties

        if self.tags:
            merged_tags = {}
            if self.copy_ami_tags:
                merged_tags.update(self.image.tags)
            merged_tags.update(self.tags)

            result["Tags"] = [{"Key": k, "Value": v} for k, v in merged_tags.items()]

        if self.terminate_on_shutdown:
            result["InstanceInitiatedShutdownBehavior"] = "terminate"

        return result


class VPC(Resource):
    """EC2 VPC."""

    ATTRIBUTES = (
        "CidrBlock",
        "CidrBlockAssociations",
        "DefaultNetworkAcl",
        "DefaultSecurityGroup",
        "Ipv6CidrBlocks",
    )

    def __init__(self, name, cidr_block):
        """Initialize a VPC.

        :param name: logical name in stack
        :type name: str
        :param cidr_block: IPv4 address range
        :type cidr_block: str
        """
        super().__init__(name, kind=AWSType.EC2_VPC)
        self.cidr_block = cidr_block

    @property
    def properties(self):
        return {
            "CidrBlock": self.cidr_block,
            "EnableDnsHostnames": True,
            "EnableDnsSupport": True,
        }

    @property
    def cidrblock(self):
        return self.getatt("CidrBlock")


class VPCEndpoint(Resource):
    """VPC Endpoint to Amazon Service."""

    def __init__(self, name, service, vpc, route_tables, policy_document):
        """Initialize a VPC endpoint.

        :param name: logical name in the stack of the entity
        :type name: str
        :param service: name of the service to connect (s3 or dynamodb)
        :type service: str
        :param vpc: VPC in which the endpoint is attached to
        :type vpc: e3.aws.cfn.ec2.VPC
        :param route_tables: a list of route table that have access to the
            endpoint.
        :type route_tables: list[RouteTable]
        :param policy_document: policy document attached to the endpoint.
        :type policy_docyment: e3.aws.cfn.ec2.security.PolicyDocument
        """
        super().__init__(name, kind=AWSType.EC2_VPC_ENDPOINT)
        assert service in ("dynamodb", "s3"), "Invalid service: %s" % service
        self.service = service
        assert isinstance(vpc, VPC), "VPC instance expected"
        self.vpc = vpc
        self.route_tables = route_tables
        for rt in self.route_tables:
            assert isinstance(rt, RouteTable), "RouteTable expected"
        self.policy_document = policy_document
        assert isinstance(self.policy_document, PolicyDocument)

    @property
    def properties(self):
        return {
            "VpcId": self.vpc.ref,
            "ServiceName": Join(
                ["com.amazonaws.", Ref("AWS::Region"), "." + self.service]
            ),
            "PolicyDocument": self.policy_document.properties,
            "RouteTableIds": [rt.ref for rt in self.route_tables],
        }


class VPCInterfaceEndpoint(Resource):
    def __init__(
        self,
        name: str,
        service: str,
        subnet: Subnet,
        vpc: VPC,
        policy_document: PolicyDocument | None,
        security_group: SecurityGroup,
    ):
        """Initialize a VPC interface endpoint.

        :param name: logical name in the stack of the entity
        :param service: name of the service to connect
        :param vpc: VPC in which the endpoint is attached to
        :param subnet: The subnet in which to create an endpoint network
            interface.
        :param policy_document: policy document attached to the endpoint.
        :param security_group: security group to associate with the endpoint
            network interface
        """
        super().__init__(name, kind=AWSType.EC2_VPC_ENDPOINT)
        self.service = service
        self.vpc = vpc
        self.subnet = subnet
        self.policy_document = policy_document
        self.security_group = security_group

    @property
    def properties(self):
        props = {
            "VpcId": self.subnet.vpc.ref,
            "ServiceName": Join(
                ["com.amazonaws.", Ref("AWS::Region"), "." + self.service]
            ),
            "PrivateDnsEnabled": "true",
            "VpcEndpointType": "Interface",
            "SubnetIds": [self.subnet.ref],
            "SecurityGroupIds": [self.security_group.group_id],
        }

        if self.policy_document is not None:
            props["PolicyDocument"] = self.policy_document.properties

        return props


class Subnet(Resource):
    """EC2 subnet."""

    def __init__(self, name, vpc, cidr_block):
        """Initialize a subnet.

        :param name: logical name in stack
        :type name: str
        :param vpc: vpc in which the subnet should be created
        :type vpc: VPC
        :param cidr_block: IPv4 address range
        :type cidr_block: str
        """
        super().__init__(name, kind=AWSType.EC2_SUBNET)
        self.cidr_block = cidr_block
        assert isinstance(vpc, VPC)
        self.vpc = vpc

    @property
    def properties(self):
        return {"CidrBlock": self.cidr_block, "VpcId": self.vpc.ref}


class InternetGateway(Resource):
    """EC2 Internet gateway."""

    def __init__(self, name):
        """Initialize an internet gateway.

        :param name: logical name in stack
        :type name: str
        """
        super().__init__(name, kind=AWSType.EC2_INTERNET_GATEWAY)


class EIP(Resource):
    """EC2 Elastic IP."""

    ATTRIBUTES = ("AllocationId",)

    def __init__(self, name, gateway_attach, instance=None):
        """Initialize Elastic IP address.

        :param name: logical name in stack
        :type name: str
        :param gateway_attach: gateway attachment
        :type gateway_attach: VPCGatewayAttachment
        :param instance: instance to which EIP is asstached
        :type instance: Instance | None
        """
        super().__init__(name, kind=AWSType.EC2_EIP)
        assert isinstance(gateway_attach, VPCGatewayAttachment)
        self.depends = gateway_attach.name
        self.instance = instance

    @property
    def allocation_id(self):
        return GetAtt(self.name, "AllocationId")

    @property
    def properties(self):
        result = {"Domain": "vpc"}
        if self.instance is not None:
            result["InstanceId"] = self.instance.ref
        return result


class NatGateway(Resource):
    """EC2 NatGateway."""

    def __init__(self, name, eip, subnet):
        """Initialize a NAT gateway.

        :param name: logical name in stack
        :type name: str
        :param eip: Elastic IP of the gateway
        :type eip: EIP
        :param subnet: subnet in which the Gateway is declared. Note that this
            should be a public subnet
        :type subnet: Subnet
        """
        super().__init__(name, kind=AWSType.EC2_NAT_GATEWAY)
        self.eip = eip
        assert isinstance(subnet, Subnet)
        self.subnet = subnet

    @property
    def properties(self):
        return {"AllocationId": self.eip.allocation_id, "SubnetId": self.subnet.ref}


class VPCGatewayAttachment(Resource):
    """EC2 VPCGatewayAttachment."""

    def __init__(self, name, vpc, gateway):
        """Initialize an attachment between a gateway and a VPC.

        :param name: logical name in stack
        :type name: str
        :param vpc: vpc in which the subnet should be created
        :type vpc: VPC
        :param gateway: a gateway
        :type gateway: InternetGateway
        """
        super().__init__(name, kind=AWSType.EC2_VPC_GATEWAY_ATTACHMENT)
        assert isinstance(vpc, VPC)
        assert isinstance(gateway, InternetGateway)
        self.vpc = vpc
        self.gateway = gateway

    @property
    def properties(self):
        return {"VpcId": self.vpc.ref, "InternetGatewayId": self.gateway.ref}


class RouteTable(Resource):
    """EC2 Route Table."""

    def __init__(self, name, vpc, tags=None):
        """Initialize a route table.

        :param name: logical name in stack
        :type name: str
        :param vpc: a VPC instance to attach the route table to.
        :type vpc: e3.aws.cfn.ec2.VPC
        :param tags: a dict of key/value tags
        :type tags: dict
        """
        super().__init__(name, kind=AWSType.EC2_ROUTE_TABLE)
        assert isinstance(vpc, VPC)
        self.vpc = vpc
        self.tags = tags

    @property
    def properties(self):
        result = {"VpcId": self.vpc.ref}
        if self.tags is not None:
            result["Tags"] = self.tags
        return result


class Route(Resource):
    """EC2 Route."""

    def __init__(self, name, route_table, dest_cidr_block, gateway, gateway_attach):
        """Initialize a route.

        :param name: logical name in stack
        :type name: str
        :param route_rable: a route table
        :type route_table: RouteTable
        :param dest_cidr_block: route ipv4 address range
        :type dest_cidr_block: str
        :param gateway: the gateway
        :type gateway: InternetGateway | NatGateway
        :param gateway_attach: a gateway attachment instance
        :type gateway_attach: VPCGatewayAttachment
        """
        super().__init__(name, kind=AWSType.EC2_ROUTE)
        assert isinstance(route_table, RouteTable)
        assert isinstance(gateway, InternetGateway) or isinstance(gateway, NatGateway)
        self.route_table = route_table
        self.dest_cidr_block = dest_cidr_block
        self.gateway = gateway
        self.gateway_attach = gateway_attach
        self.depends = self.gateway_attach.name

    @property
    def properties(self):
        result = {
            "RouteTableId": self.route_table.ref,
            "DestinationCidrBlock": self.dest_cidr_block,
        }
        if isinstance(self.gateway, InternetGateway):
            result["GatewayId"] = self.gateway.ref
        else:
            result["NatGatewayId"] = self.gateway.ref
        return result


class SubnetRouteTableAssociation(Resource):
    """EC2 SubnetRouteTableAssociation."""

    def __init__(self, name, subnet, route_table):
        """Initialize an association between a route table and a subnet.

        :param name: logical name in stack
        :type name: str
        :param subnet: a subnet instance to attach the route table to.
        :type subnet: e3.aws.cfn.ec2.Subnet
        :param route_rable: a route table
        :type route_table: RouteTable
        """
        super().__init__(name, kind=AWSType.EC2_SUBNET_ROUTE_TABLE_ASSOCIATION)
        self.subnet = subnet
        self.route_table = route_table

    @property
    def properties(self):
        return {"SubnetId": self.subnet.ref, "RouteTableId": self.route_table.ref}



from __future__ import annotations
from dataclasses import dataclass, field
from typing import TYPE_CHECKING


from troposphere import AWSObject, ecs, Ref, Tags

from e3.aws import name_to_id
from e3.aws.troposphere import Construct

if TYPE_CHECKING:
    from e3.aws.troposphere import Stack


@dataclass(frozen=True)
class FargateTaskDefinition(Construct):
    """Define a Fargate Task.

    :param name: name of the task definition
    :param container_definition: a list of troposphere container definitions
        that make up your task
    :param cpu: number of cpu units used by the task
    :param memory: amount (in MiB) of memory used by the task
    :param family: name of a family that this task definition is registered to
    :param tags: the metadata that you apply to the task definition to help you
        categorize and organize them
    :param task_role_arn: the short name or full Amazon Resource Name (ARN) of
        the AWS Identity and Access Management (IAM) role that grants containers
        in the task permission to call AWS APIs on your behalf.
    :param volumes: the list of volume definitions for the task
    """

    name: str
    container_definitions: list[ecs.ContainerDefintion]

    cpu: str = "256"
    memory: str = "512"

    family: str | None = None
    tags: dict[str, str] = field(default_factory=lambda: {})
    task_role_arn: str | None = None
    volumes: list[ecs.Volume] | None = None

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Construct and return Fargate TaskDefinition resources."""
        kwargs = {
            key: val
            for key, val in {
                "ContainerDefinitions": self.container_definitions,
                "Cpu": self.cpu,
                "Memory": self.memory,
                "Family": self.family,
                "NetworkMode": "awsvpc",
                "RequiresCompatibilities": ["FARGATE"],
                "ExecutionRoleArn": Ref(name_to_id("ECSTaskExecutionRole")),
                "TaskRoleArn": self.task_role_arn,
                "Tags": Tags({"Name": self.name, **self.tags}),
                "Volumes": self.volumes,
            }.items()
            if val is not None
        }

        return [ecs.TaskDefinition(name_to_id(self.name), **kwargs)]



from e3.aws.cfn.codecommit import Repository


def test_create_repository():
    """RecordSet test."""
    name = "my-repo"
    description = "my repo description"
    r = Repository(name, description)
    assert r.properties["RepositoryName"] == name
    assert r.properties["RepositoryDescription"] == description

    # Resource name should only include alphanum characters
    assert r.name == "myrepo"



from __future__ import absolute_import, division, print_function

import pytest
import yaml
from botocore.stub import ANY
from e3.aws import AWSEnv, default_region
from e3.aws.cfn import Stack
from e3.aws.cfn.ec2 import (
    EIP,
    VPC,
    EBSDisk,
    EC2NetworkInterface,
    EphemeralDisk,
    Instance,
    InternetGateway,
    NatGateway,
    Route,
    RouteTable,
    Subnet,
    SubnetRouteTableAssociation,
    UserData,
    VPCEndpoint,
    VPCGatewayAttachment,
    WinUserData,
)
from e3.aws.cfn.ec2.security import SecurityGroup
from e3.aws.cfn.iam import Allow, PolicyDocument, Principal, PrincipalKind
from e3.aws.ec2.ami import AMI


def test_create_network():
    s = Stack(name="teststack")

    s = Stack(name="MyStack")
    s += VPC("BuildVPC", "10.10.0.0/16")
    s += InternetGateway("Gate")
    s += Subnet("BuildPublicSubnet", s["BuildVPC"], "10.10.10.0/24")
    s += Subnet("BuildPrivateSubnet", s["BuildVPC"], "10.10.20.0/24")
    s += VPCGatewayAttachment("GateAttach", s["BuildVPC"], s["Gate"])
    s += RouteTable("RT", s["BuildVPC"])
    s += Route("PRoute", s["RT"], "0.0.0.0/0", s["Gate"], s["GateAttach"])
    s += SubnetRouteTableAssociation("RTSAssoc", s["BuildPublicSubnet"], s["RT"])
    p = PolicyDocument().append(
        Allow(
            to="GetObject",
            on="arn:aws:s3:::abucket/*",
            apply_to=Principal(PrincipalKind.SERVICE, "ec2.amazonaws.com"),
        )
    )

    s += VPCEndpoint("S3EndPoint", "s3", s["BuildVPC"], [s["RT"]], policy_document=p)
    assert s.body


def test_create_instance():
    aws_env = AWSEnv(regions=["us-east-1"], stub=True)
    with default_region("us-east-1"):
        stub = aws_env.stub("ec2", region="us-east-1")
        stub.add_response(
            "describe_images",
            {
                "Images": [
                    {"ImageId": "ami-1234", "RootDeviceName": "/dev/sda1", "Tags": []}
                ]
            },
            {"ImageIds": ANY},
        )

        i = Instance("testmachine", AMI("ami-1234"), disk_size=20)
        assert i.properties

        i.add(EphemeralDisk("/dev/sdb", 0))
        assert i.properties

        i.add(EBSDisk("/dev/sdc", size=20, encrypted=True))
        assert i.properties

        vpc = VPC("VPC", "10.10.0.0/16")
        subnet = Subnet("Subnet", vpc, "10.10.10.0/24")
        subnet = Subnet("Subnet2", vpc, "10.10.20.0/24")
        security_group = SecurityGroup("mysgroup", vpc)
        i.add(EC2NetworkInterface(subnet, description="first network interface"))
        i.add(
            EC2NetworkInterface(
                subnet, groups=[security_group], description="2nd network interface"
            )
        )
        i.add(
            EC2NetworkInterface(
                subnet,
                groups=[security_group],
                description="3rd network interface",
                device_index=3,
            )
        )
        assert i.properties

        with pytest.raises(AssertionError):
            i.add("non valid ec2 device")


def test_user_data_creation():
    """Test creation of user data."""
    a = UserData()
    a.add("toto.txt", "x-shellscript", "Hello")
    a.add("url1", "x-include-url", "toto.jpg")
    assert yaml.dump(a.properties)

    aws_env = AWSEnv(regions=["us-east-1"], stub=True)
    with default_region("us-east-1"):
        stub = aws_env.stub("ec2", region="us-east-1")
        stub.add_response(
            "describe_images",
            {
                "Images": [
                    {"ImageId": "ami-1234", "RootDeviceName": "/dev/sda1", "Tags": []}
                ]
            },
            {"ImageIds": ANY},
        )

        i = Instance("testmachine", AMI("ami-1234"))
        i.add_user_data("url1", "x-include-url", "http://dummy")
        assert i.properties


def test_win_user_data_creation():
    """Test creation of windows user data."""
    a = WinUserData()
    a.add("powershell", 'echo "test powezrshell"')
    a.add("script", "echo test script")
    assert yaml.dump(a.properties)

    aws_env = AWSEnv(regions=["us-east-1"], stub=True)
    with default_region("us-east-1"):
        stub = aws_env.stub("ec2", region="us-east-1")
        stub.add_response(
            "describe_images",
            {
                "Images": [
                    {
                        "ImageId": "ami-1234-win",
                        "RootDeviceName": "/dev/sda1",
                        "Platform": "windows",
                        "Tags": [],
                    }
                ]
            },
            {"ImageIds": ANY},
        )

        i = Instance("testmachine", AMI("ami-1234-win"))
        i.add_user_data("persist", "true")
        assert i.properties


def test_cfn_init_set():
    s = Stack(name="teststack")

    aws_env = AWSEnv(regions=["us-east-1"], stub=True)
    with default_region("us-east-1"):
        stub = aws_env.stub("ec2", region="us-east-1")
        stub.add_response(
            "describe_images",
            {
                "Images": [
                    {"ImageId": "ami-1234", "RootDeviceName": "/dev/sda1", "Tags": []}
                ]
            },
            {"ImageIds": ANY},
        )

        s += Instance("server", AMI("ami-1234"))
        s["server"].set_cfn_init()
        assert s.body


def test_nat_gateway():
    """Create a NATGateway."""
    s = Stack(name="MyStack")
    s += VPC("BuildVPC", "10.10.0.0/16")
    s += Subnet("BuildPublicSubnet", s["BuildVPC"], "10.10.10.0/24")
    s += Subnet("BuildPrivateSubnet", s["BuildVPC"], "10.10.20.0/24")
    s += InternetGateway("Gate")
    s += VPCGatewayAttachment("GateAttach", s["BuildVPC"], s["Gate"])
    s += RouteTable("RT", s["BuildVPC"])
    s += Route("PRoute", s["RT"], "0.0.0.0/0", s["Gate"], s["GateAttach"])
    s += SubnetRouteTableAssociation("RTSAssoc", s["BuildPublicSubnet"], s["RT"])
    s += EIP("NatEip", s["GateAttach"])
    s += NatGateway("NatGate", s["NatEip"], s["BuildPublicSubnet"])

    s += RouteTable("NATRT", s["BuildVPC"])
    s += Route("NATRoute", s["NATRT"], "0.0.0.0/0", s["NatGate"], s["GateAttach"])
    s += SubnetRouteTableAssociation("NatRTSAssoc", s["BuildPrivateSubnet"], s["NATRT"])
    assert s.body



"""Provide S3 construct tests."""
import json
import os

from e3.aws.troposphere.s3.bucket import Bucket, EncryptionAlgorithm
from e3.aws.troposphere.s3 import BucketWithRoles
from e3.aws.troposphere import Stack
from e3.aws.troposphere.awslambda import Py38Function
from e3.aws.troposphere.sns import Topic
from e3.aws.troposphere.sqs import Queue


TEST_DIR = os.path.dirname(os.path.abspath(__file__))


def test_bucket(stack: Stack) -> None:
    """Test bucket creation."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"

    topic_test = Topic(name="test-topic")
    queue_test = Queue(name="test-queue")
    lambda_test = Py38Function(
        name="mypylambda",
        description="this is a test",
        role="somearn",
        code_dir="my_code_dir",
        handler="app.main",
    )

    stack.add(topic_test)
    stack.add(lambda_test)
    stack.add(queue_test)

    bucket = Bucket(name="test-bucket")
    bucket.add_notification_configuration(
        event="s3:ObjectCreated:*", target=topic_test, permission_suffix="TpUpload"
    )
    bucket.add_notification_configuration(
        event="s3:ObjectCreated:*", target=lambda_test, permission_suffix="TpUpload"
    )
    bucket.add_notification_configuration(
        event="s3:ObjectCreated:*", target=queue_test, permission_suffix="FileEvent"
    )
    stack.add(bucket)

    with open(os.path.join(TEST_DIR, "bucket.json")) as fd:
        expected_template = json.load(fd)

    assert stack.export()["Resources"] == expected_template


def test_bucket_with_roles(stack: Stack) -> None:
    """Test BucketWithRoles."""
    bucket = BucketWithRoles(
        name="test-bucket-with-roles",
        iam_names_prefix="TestBucket",
        iam_read_root_name="Restore",
        iam_write_root_name="Push",
        iam_path="/test/",
        trusted_accounts=["123456789"],
    )
    stack.add(bucket)

    with open(os.path.join(TEST_DIR, "bucket-with-roles.json")) as fd:
        expected_template = json.load(fd)

    assert stack.export()["Resources"] == expected_template


def test_bucket_multi_encryption(stack: Stack) -> None:
    """Test bucket accepting multiple types of encryptions and without default."""
    bucket = Bucket(
        name="test-bucket",
        default_bucket_encryption=None,
        authorized_encryptions=[EncryptionAlgorithm.AES256, EncryptionAlgorithm.KMS],
    )
    stack.add(bucket)

    with open(os.path.join(TEST_DIR, "bucket_multi_encryption.json")) as fd:
        expected_template = json.load(fd)

    assert stack.export()["Resources"] == expected_template


def test_bucket_notification_string_arns(stack: Stack) -> None:
    """Test bucket notification with string arns instead of objects."""
    bucket = Bucket(name="test-bucket")
    bucket.add_notification_configuration(
        event="s3:ObjectCreated:*",
        target="arn:aws:sns:us-east-2:123456789012:MyTopic",
        permission_suffix="TpUpload",
    )
    bucket.add_notification_configuration(
        event="s3:ObjectCreated:*",
        target="arn:aws:lambda:us-east-2:123456789012:MyFunction",
        permission_suffix="TpUpload",
    )
    bucket.add_notification_configuration(
        event="s3:ObjectCreated:*",
        target="arn:aws:sqs:us-east-2:123456789012:MyQueue",
        permission_suffix="FileEvent",
    )
    stack.add(bucket)

    with open(os.path.join(TEST_DIR, "bucket_notification_string_arns.json")) as fd:
        expected_template = json.load(fd)

    assert stack.export()["Resources"] == expected_template



from __future__ import absolute_import, division, print_function

import pytest

from botocore.stub import ANY
from e3.aws import AWSEnv, default_region
from e3.aws.cfn.arch import Fortress, AWSFortressError
from e3.aws.cfn.iam import Allow, Policy, PolicyDocument
from e3.aws.cfn.s3 import Bucket
from e3.aws.ec2.ami import AMI
from e3.aws.cfn.ec2.security import SecurityGroup

AWS_IP_RANGES = {
    "syncToken": "1600978876",
    "createDate": "2020-09-24-20-21-16",
    "prefixes": [
        {
            "ip_prefix": "3.5.140.0/22",
            "region": "ap-northeast-2",
            "service": "AMAZON",
            "network_border_group": "ap-northeast-2",
        },
        {
            "ip_prefix": "120.52.22.96/27",
            "region": "GLOBAL",
            "service": "AMAZON",
            "network_border_group": "GLOBAL",
        },
        {
            "ip_prefix": "150.222.81.0/24",
            "region": "eu-west-1",
            "service": "AMAZON",
            "network_border_group": "eu-west-1",
        },
        {
            "ip_prefix": "52.4.0.0/14",
            "region": "us-east-1",
            "service": "AMAZON",
            "network_border_group": "us-east-1",
        },
    ],
}

GITHUB_API_RANGE = {"git": ["127.0.0.0/24", "2a0a:a440::/29"]}


@pytest.mark.parametrize("enable_github", [True, False])
def test_create_fortress(enable_github, requests_mock):
    if enable_github:
        requests_mock.get("https://api.github.com/meta", json=GITHUB_API_RANGE)
    requests_mock.get(
        "https://ip-ranges.amazonaws.com/ip-ranges.json", json=AWS_IP_RANGES
    )
    aws_env = AWSEnv(regions=["us-east-1"], stub=True)
    with default_region("us-east-1"):
        stub = aws_env.stub("ec2", region="us-east-1")
        stub.add_response(
            "describe_images",
            {
                "Images": [
                    {"ImageId": "ami-1234", "RootDeviceName": "/dev/sda1", "Tags": []}
                ]
            },
            {"ImageIds": ANY},
        )
        stub.add_response(
            "describe_images",
            {
                "Images": [
                    {"ImageId": "ami-1234", "RootDeviceName": "/dev/sda1", "Tags": []}
                ]
            },
            {"ImageIds": ANY},
        )
        d = PolicyDocument().append(
            Allow(
                to="s3:GetObject",
                on=["arn:aws:s3:::mybucket", "arn:aws:s3:::mybucket/*"],
            )
        )
        p = Policy("InternalPolicy", d)
        f = Fortress(
            "myfortress",
            allow_ssh_from="0.0.0.0/0",
            bastion_ami=AMI("ami-1234"),
            internal_server_policy=p,
        )
        f += Bucket("Bucket2")

        # Allow access to mybucket through a s3 endpoint
        f.private_subnet.add_bucket_access(["mybucket", f["Bucket2"]])

        # Allow access to a secret throught a secretsmanager endpoint
        f.add_secret_access("arn_secret")

        # Allow access to lambdas throught lambda endpoints
        f.add_lambda_access(["arn_lambda_1", "arn_lambda_2"])

        # Allow access to sts service
        f.add_service_access("sts")

        # allow https
        f.add_network_access("https")
        f.add_private_server(
            AMI("ami-1234"), ["server1", "server2"], github_access=enable_github
        )

        assert f.body


def test_create_fortress_no_bastion():
    aws_env = AWSEnv(regions=["us-east-1"], stub=True)
    with default_region("us-east-1"):
        stub = aws_env.stub("ec2", region="us-east-1")
        stub.add_response(
            "describe_images",
            {
                "Images": [
                    {"ImageId": "ami-1234", "RootDeviceName": "/dev/sda1", "Tags": []}
                ]
            },
            {"ImageIds": ANY},
        )
        stub.add_response(
            "describe_images",
            {
                "Images": [
                    {"ImageId": "ami-1234", "RootDeviceName": "/dev/sda1", "Tags": []}
                ]
            },
            {"ImageIds": ANY},
        )
        d = PolicyDocument().append(
            Allow(
                to="s3:GetObject",
                on=["arn:aws:s3:::mybucket", "arn:aws:s3:::mybucket/*"],
            )
        )
        p = Policy("InternalPolicy", d)
        f = Fortress("myfortress", bastion_ami=None, internal_server_policy=p)
        f += Bucket("Bucket2")

        # Allow access to mybucket through a s3 endpoint
        f.private_subnet.add_bucket_access(["mybucket", f["Bucket2"]])

        # Allow access to a secret throught a secretsmanager endpoint
        f.add_secret_access("arn_secret")

        # allow https
        f.add_network_access("https")
        f.add_private_server(AMI("ami-1234"), ["server1", "server2"])

        assert f.body


def test_create_fortress_with_too_much_sgs():
    aws_env = AWSEnv(regions=["us-east-1"], stub=True)
    with default_region("us-east-1"):
        stub = aws_env.stub("ec2", region="us-east-1")
        stub.add_response(
            "describe_images",
            {
                "Images": [
                    {"ImageId": "ami-1234", "RootDeviceName": "/dev/sda1", "Tags": []}
                ]
            },
            {"ImageIds": ANY},
        )

        d = PolicyDocument().append(
            Allow(
                to="s3:GetObject",
                on=["arn:aws:s3:::mybucket", "arn:aws:s3:::mybucket/*"],
            )
        )
        p = Policy("InternalPolicy", d)
        f = Fortress("myfortress", bastion_ami=None, internal_server_policy=p)

        # Adding 16 extra security groups should raise an exception (The maximum
        # number of security groups is 16 and there is a default InternalSG)
        sg_groups = [SecurityGroup(name=f"sg{id}", vpc=f.vpc.vpc) for id in range(16)]
        with pytest.raises(AWSFortressError):
            f.add_private_server(
                AMI("ami-1234"),
                ["server1"],
                amazon_access=False,
                github_access=False,
                extra_groups=sg_groups,
            )



"""Provide PolicyStatement class."""

from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Union, List, Dict

    ResourceType = Union[str, List[str]]
    PrincipalType = Union[str, Dict[str, Union[str, List[str]]]]
    ConditionType = Dict[str, Dict[str, Union[str, List[str]]]]


class PolicyStatement:
    """Default Policy statement class."""

    def __init__(
        self,
        action: str | list[str],
        effect: str = "Deny",
        resource: ResourceType | None = None,
        principal: PrincipalType | None = None,
        condition: ConditionType | None = None,
    ) -> None:
        """Initialize a policy statement.

        :param action: actions on which the policy has effect
        :param effect: effect of the policy (Allow, Deny ..)
        :param resource: resource on which the policy has effect
        :param principal: principal affected by the policy
        :param condition: conditions for when the policy is in effect
        """
        self.action = action
        self.effect = effect
        self.resource = resource
        self.principal = principal
        self.condition = condition

    @property
    def as_dict(self) -> dict[str, Any]:
        """Return a dictionnary defining a troposphere policy statement."""
        return {
            key: val
            for key, val in {
                "Effect": self.effect,
                "Principal": self.principal,
                "Action": self.action,
                "Resource": self.resource,
                "Condition": self.condition,
            }.items()
            if val is not None
        }


class Allow(PolicyStatement):
    def __init__(
        self,
        action: str | list[str],
        resource: ResourceType | None = None,
        principal: PrincipalType | None = None,
        condition: ConditionType | None = None,
    ) -> None:
        """Initialize an Allow policy statement.

        :param action: actions on which the policy has effect
        :param resource: resource on which the policy has effect
        :param principal: principal affected by the policy
        :param condition: conditions for when the policy is in effect
        """
        return super().__init__(
            action=action,
            effect="Allow",
            resource=resource,
            principal=principal,
            condition=condition,
        )


class AssumeRole(PolicyStatement):
    """Define a sts:AssumeRole role policy statement."""

    def __init__(
        self, principal: PrincipalType, condition: ConditionType | None = None
    ):
        """Initialize an AssumeRole statement.

        :param principal: principal which are allowed to assume the role
        :param condition: condition to apply
        """
        super().__init__(
            action="sts:AssumeRole",
            effect="Allow",
            resource=None,
            principal=principal,
            condition=condition,
        )


class Trust(PolicyStatement):
    """Policy statement used in trust policies."""

    def __init__(
        self,
        services: list[str] | None = None,
        accounts: list[str] | None = None,
        users: list[tuple[str, str]] | None = None,
        roles: list[tuple[str, str]] | None = None,
        condition: ConditionType | None = None,
        actions: list[str] | str = "sts:AssumeRole",
    ) -> None:
        """Initialize a trust policy statement.

        :param services: list of services to trust (without amazonaws.com suffix)
        :param accounts: list of accounts to trust (accounts alias not allowed)
        :param users: list of users as tuple (account number, user name)
        :param roles: list of roles as tuple (account number, role name)
        :param condition: condition to apply to the statement
        :param actions: list of trusted actions. If None, the action iam:AssumeRole
            is used.
        """
        self.principals: dict[str, list[str]] = {}

        if services is not None:
            self.principals.setdefault("Service", [])
            self.principals["Service"] += [
                f"{service}.amazonaws.com" for service in services
            ]

        if accounts is not None:
            self.principals.setdefault("AWS", [])
            self.principals["AWS"] += [
                f"arn:aws:iam::{account}:root" for account in accounts
            ]

        if users is not None:
            self.principals.setdefault("AWS", [])
            self.principals["AWS"] += [
                f"arn:aws:iam::{account}:user/{user}" for account, user in users
            ]

        if roles is not None:
            self.principals.setdefault("AWS", [])
            self.principals["AWS"] += [
                f"arn:aws:iam::{account}:role/{role}" for account, role in roles
            ]

        self.condition = condition
        self.actions = actions

    @property
    def as_dict(self) -> dict[str, Any]:
        """See PolicyStatement doc."""
        result: dict[str, str | list[str] | dict[str, list[str]] | ConditionType] = {
            "Effect": "Allow",
            "Action": self.actions,
            "Principal": self.principals,
        }

        if self.condition is not None:
            result["Condition"] = self.condition

        return result



"""Provide Stack tests."""

from e3.aws.troposphere.s3.bucket import Bucket
from e3.aws.troposphere import Stack


def test_instanciate() -> None:
    """Test stack instanciation."""
    stack = Stack("test-stack", "this is a test stack")
    assert stack


def test_add_and_get_item() -> None:
    """Test adding a construct and retrieving an AWSObject from a stack."""
    stack = Stack("test-stack", "this is a test stack")
    stack.add(Bucket("my-bucket"))
    my_bucket = stack["my-bucket"]
    assert my_bucket



"""Provide Cloudformation construct tests."""
from __future__ import annotations
import json
import os

from e3.aws.troposphere import Stack
from e3.aws.troposphere.s3 import Bucket
from e3.aws.troposphere.cloudfront import S3WebsiteDistribution


TEST_DIR = os.path.dirname(os.path.abspath(__file__))


def test_s3_website_distribution(stack: Stack) -> None:
    """Test Cloudfront S3WebsiteDistribution construct."""
    stack.add(
        S3WebsiteDistribution(
            name="test-s3w-dist",
            aliases=["test.s3w.com"],
            bucket_name="host-bucket",
            certificate_arn="acm_arn",
            default_ttl=360,
            lambda_edge_function_arns=["lamba_arn"],
            r53_route_from=[("hosted_zone_id", "test.s3w.com")],
        )
    )

    with open(os.path.join(TEST_DIR, "s3websitedistribution.json")) as fd:
        expected_template = json.load(fd)

    assert stack.export()["Resources"] == expected_template


def test_s3_website_distribution_logging_default(stack: Stack) -> None:
    """Test Cloudfront S3WebsiteDistribution construct with default logging."""
    bucket = Bucket(name="test-bucket")

    stack.add(
        S3WebsiteDistribution(
            name="test-s3w-dist",
            aliases=["test.s3w.com"],
            bucket_name="host-bucket",
            certificate_arn="acm_arn",
            default_ttl=360,
            lambda_edge_function_arns=["lamba_arn"],
            r53_route_from=[("hosted_zone_id", "test.s3w.com")],
            logging_bucket=bucket.domain_name,
        )
    )

    with open(
        os.path.join(TEST_DIR, "s3websitedistribution_logging_default.json")
    ) as fd:
        expected_template = json.load(fd)

    assert stack.export()["Resources"] == expected_template


def test_s3_website_distribution_logging(stack: Stack) -> None:
    """Test Cloudfront S3WebsiteDistribution construct with logging."""
    bucket = Bucket(name="test-bucket")

    stack.add(
        S3WebsiteDistribution(
            name="test-s3w-dist",
            aliases=["test.s3w.com"],
            bucket_name="host-bucket",
            certificate_arn="acm_arn",
            default_ttl=360,
            lambda_edge_function_arns=["lamba_arn"],
            r53_route_from=[("hosted_zone_id", "test.s3w.com")],
            logging_bucket=bucket.regional_domain_name,
            logging_prefix="myprefix",
            logging_include_cookies=True,
        )
    )

    with open(os.path.join(TEST_DIR, "s3websitedistribution_logging.json")) as fd:
        expected_template = json.load(fd)

    assert stack.export()["Resources"] == expected_template



from __future__ import annotations, absolute_import, division, print_function

from datetime import datetime
import os
import pytest
from typing import TYPE_CHECKING

from botocore.stub import ANY
from e3.aws import AWSEnv, default_region
from e3.aws.cfn import Stack
from e3.aws.cfn.main import CFNMain

if TYPE_CHECKING:
    from _pytest.monkeypatch import MonkeyPatch

DEFAULT_S3_ANSWER = {
    "ResponseMetadata": {"HTTPStatusCode": 200, "RetryAttempts": 1},
    "ETag": '"f71dbe52628a3eeee3a77ab494817525c6"',
    "VersionId": "AQSVsaeeeeeeY1r95GEBkqOyubejeVl",
}


def test_cfn_main() -> None:
    class MyCFNMain(CFNMain):
        def create_stack(self) -> Stack:
            return Stack(name="teststack")

    aws_env = AWSEnv(regions=["us-east-1"], stub=True)
    with default_region("us-east-1"):
        aws_env.client("cloudformation", region="us-east-1")

        stubber = aws_env.stub("cloudformation")
        stubber.add_response("validate_template", {}, {"TemplateBody": ANY})
        stubber.add_response(
            "create_stack",
            {},
            {
                "Capabilities": ["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM"],
                "StackName": "teststack",
                "ClientRequestToken": ANY,
                "TemplateBody": ANY,
            },
        )
        stubber.add_response("describe_stacks", {}, {"StackName": "teststack"})
        with stubber:
            m = MyCFNMain(regions=["us-east-1"])
            m.execute(args=["push", "--no-wait"], aws_env=aws_env)


def test_cfn_main_multiple_stacks() -> None:
    class MyCFNMain(CFNMain):
        def create_stack(self) -> list[Stack]:
            return [Stack(name="first-stack"), Stack(name="second-stack")]

    aws_env = AWSEnv(regions=["us-east-1"], stub=True)
    with default_region("us-east-1"):
        aws_env.client("cloudformation", region="us-east-1")

        stubber = aws_env.stub("cloudformation")
        for stack_name in ("first-stack", "second-stack"):
            stubber.add_response("validate_template", {}, {"TemplateBody": ANY})
            stubber.add_response(
                "create_stack",
                {},
                {
                    "Capabilities": ["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM"],
                    "StackName": f"{stack_name}",
                    "ClientRequestToken": ANY,
                    "TemplateBody": ANY,
                },
            )
        with stubber:
            m = MyCFNMain(regions=["us-east-1"])
            m.execute(args=["push", "--no-wait"], aws_env=aws_env)


@pytest.mark.parametrize(
    "status",
    [
        ("NOT_FAILED", "Other Reason", 0),
        (
            "FAILED",
            (
                "The submitted information didn't contain changes."
                "Submit different information to create a change set."
            ),
            0,
        ),
        ("FAILED", "Other Reason", 1),
    ],
)
def test_cfn_main_push_existing_stack(
    status: tuple[str, str, int], monkeypatch: MonkeyPatch
) -> None:
    """Test pushing an already existing stack.

    :param status: Tuple with status and status reason from describe_change_set
        response and associated expected execute return value.
    """

    class MyCFNMain(CFNMain):
        def create_stack(self):
            return [Stack(name="existing-stack")]

    aws_env = AWSEnv(regions=["us-east-1"], stub=True)

    with default_region("us-east-1"):
        aws_env.client("cloudformation", region="us-east-1")

        stack_name = "existing-stack"
        stubber = aws_env.stub("cloudformation")
        stubber.add_response("validate_template", {}, {"TemplateBody": ANY})
        stubber.add_response(
            "describe_stacks",
            service_response={
                "Stacks": [
                    {
                        "StackName": stack_name,
                        "CreationTime": datetime(2016, 1, 20, 22, 9),
                        "StackStatus": "CREATE_COMPLETE",
                        "StackId": stack_name + "1",
                    }
                ]
            },
            expected_params={"StackName": stack_name},
        )
        stubber.add_response(
            "create_change_set",
            {},
            {
                "Capabilities": ["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM"],
                "ChangeSetName": ANY,
                "StackName": stack_name,
                "TemplateBody": ANY,
            },
        )
        stubber = stubber
        stubber.add_response(
            "describe_change_set",
            {
                "StackName": stack_name,
                "Status": status[0],
                "StatusReason": status[1],
                "Changes": [],
            },
            {"ChangeSetName": ANY, "StackName": stack_name},
        )

        if status[0] == "FAILED":
            stubber.add_response(
                "delete_change_set", {}, {"ChangeSetName": ANY, "StackName": stack_name}
            )
        else:
            stubber.add_response(
                "execute_change_set",
                {},
                {"ChangeSetName": ANY, "StackName": ANY, "ClientRequestToken": ANY},
            )
            stubber.add_response(
                "describe_stacks",
                service_response={
                    "Stacks": [
                        {
                            "StackName": stack_name,
                            "CreationTime": datetime(2016, 1, 20, 22, 9),
                            "StackStatus": "UPDATE_COMPLETE",
                            "StackId": stack_name + "1",
                        }
                    ]
                },
                expected_params={"StackName": ANY},
            )
            monkeypatch.setattr("builtins.input", lambda _: "Y")

        with stubber:
            m = MyCFNMain(regions=["us-east-1"])
            assert m.execute(args=["update", "--no-wait"], aws_env=aws_env) == status[2]


def test_cfn_main_s3() -> None:
    class MyCFNMain(CFNMain):
        def create_stack(self) -> Stack:
            return Stack(name="teststack")

    os.mkdir("data")
    aws_env = AWSEnv(regions=["us-east-1"], stub=True)
    with default_region("us-east-1"):
        aws_env.client("cloudformation", region="us-east-1")

        stubber = aws_env.stub("cloudformation")
        stubber.add_response("validate_template", {}, {"TemplateURL": ANY})
        stubber.add_response(
            "create_stack",
            {},
            {
                "Capabilities": ["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM"],
                "StackName": "teststack",
                "ClientRequestToken": ANY,
                "TemplateURL": ANY,
            },
        )
        stubber.add_response("describe_stacks", {}, {"StackName": "teststack"})

        aws_env.client("s3", region="us-east-1")
        s3_stubber = aws_env.stub("s3")
        s3_stubber.add_response(
            "put_object",
            DEFAULT_S3_ANSWER,
            {
                "Bucket": "superbucket",
                "Body": ANY,
                "Key": ANY,
                "ServerSideEncryption": "AES256",
            },
        )
        with stubber:
            with s3_stubber:
                m = MyCFNMain(
                    regions=["us-east-1"],
                    data_dir="data",
                    s3_bucket="superbucket",
                    s3_key="test_key",
                )
                m.execute(args=["push", "--no-wait"], aws_env=aws_env)



import pytest
from e3.aws import AWSEnv, default_region, Session
from e3.aws.ec2.ami import AMI


def test_ls_ami():
    """List AMIS from all regions."""
    aws_env = AWSEnv(regions=["us-east-1", "eu-west-1"], stub=True)
    stub_us = aws_env.stub("ec2", region="us-east-1")
    stub_eu = aws_env.stub("ec2", region="eu-west-1")

    stub_eu.add_response(
        "describe_images",
        {
            "Images": [
                {"ImageId": "ami-1234", "RootDeviceName": "/dev/sda1", "Tags": []}
            ]
        },
        {"Filters": [], "Owners": ["self"]},
    )
    stub_us.add_response(
        "describe_images",
        {
            "Images": [
                {"ImageId": "ami-5678", "RootDeviceName": "/dev/sda1", "Tags": []}
            ]
        },
        {"Filters": [], "Owners": ["self"]},
    )
    assert len(AMI.ls()) == 2


def test_select():
    """Test AMI.select."""
    aws_env = AWSEnv(regions=["us-east-1"], stub=True)
    stub = aws_env.stub("ec2", region="us-east-1")

    Images = {
        "Images": [
            {
                "ImageId": "ami-1",
                "RootDeviceName": "/dev/sda1",
                "Tags": [
                    {"Key": "platform", "Value": "x86_64-linux"},
                    {"Key": "os_version", "Value": "suse11"},
                    {"Key": "timestamp", "Value": "4"},
                ],
            },
            {
                "ImageId": "ami-2",
                "RootDeviceName": "/dev/sda1",
                "Tags": [
                    {"Key": "platform", "Value": "x86_64-linux"},
                    {"Key": "os_version", "Value": "suse11"},
                    {"Key": "timestamp", "Value": "5"},
                ],
            },
            {
                "ImageId": "ami-3",
                "RootDeviceName": "/dev/sda1",
                "Tags": [
                    {"Key": "platform", "Value": "x86_64-linux"},
                    {"Key": "os_version", "Value": "suse11"},
                    {"Key": "timestamp", "Value": "1"},
                ],
            },
            {
                "ImageId": "ami-4",
                "RootDeviceName": "/dev/sda1",
                "Tags": [
                    {"Key": "platform", "Value": "x86_64-linux"},
                    {"Key": "timestamp", "Value": "1"},
                ],
            },
            {
                "ImageId": "ami-1234",
                "RootDeviceName": "/dev/sda1",
                "Tags": [
                    {"Key": "platform", "Value": "x86_64-linux"},
                    {"Key": "os_version", "Value": "ubuntu16.04"},
                    {"Key": "timestamp", "Value": "1"},
                ],
            },
            {
                "ImageId": "ami-5",
                "RootDeviceName": "/dev/sda1",
                "Tags": [
                    {"Key": "platform", "Value": "x86_64-linux"},
                    {"Key": "os_version", "Value": "suse11"},
                    {"Key": "timestamp", "Value": "5"},
                    {"Key": "kind", "Value": "build"},
                ],
            },
        ]
    }

    stub.add_response(
        "describe_images",
        Images,
        {
            "Filters": [
                {"Name": "tag-key", "Values": ["platform"]},
                {"Name": "tag-key", "Values": ["timestamp"]},
                {"Name": "tag-key", "Values": ["os_version"]},
            ],
            "Owners": ["self"],
        },
    )
    stub.add_response(
        "describe_images",
        Images,
        {
            "Filters": [
                {"Name": "tag-key", "Values": ["platform"]},
                {"Name": "tag-key", "Values": ["timestamp"]},
                {"Name": "tag-key", "Values": ["os_version"]},
                {"Name": "tag-key", "Values": ["kind"]},
            ],
            "Owners": ["self"],
        },
    )

    with default_region("us-east-1"):
        ami = AMI.select(platform="x86_64-linux", os_version="suse11")
        assert ami.id == "ami-2"
        ami = AMI.select(platform="x86_64-linux", os_version="suse11", kind="build")
        assert ami.id == "ami-5"


def test_session_without_args():
    """Raise error when no arguments."""
    with pytest.raises(ValueError):
        session = Session()
        session.client("ec2")



from __future__ import annotations

from datetime import datetime
import logging
import os
import sys
from typing import TYPE_CHECKING

from e3.archive import create_archive
from e3.fs import sync_tree, rm
from e3.os.process import Run
from troposphere import awslambda, logs, GetAtt, Ref, Sub

from e3.aws import name_to_id
from e3.aws.troposphere import Construct
from e3.aws.troposphere.iam.policy_document import PolicyDocument
from e3.aws.troposphere.iam.policy_statement import PolicyStatement
from e3.aws.troposphere.iam.role import Role
from e3.aws.util.ecr import build_and_push_image

if TYPE_CHECKING:
    from typing import Any
    from troposphere import AWSObject
    from e3.aws.troposphere import Stack

logger = logging.getLogger("e3.aws.troposphere.awslambda")


class Function(Construct):
    """A lambda function."""

    def __init__(
        self,
        name: str,
        description: str,
        role: str | GetAtt | Role,
        code_bucket: str | None = None,
        code_key: str | None = None,
        code_zipfile: str | None = None,
        handler: str | None = None,
        code_version: int | None = None,
        timeout: int = 3,
        runtime: str | None = None,
        memory_size: int | None = None,
        ephemeral_storage_size: int | None = None,
        logs_retention_in_days: int | None = 731,
        reserved_concurrent_executions: int | None = None,
        environment: dict[str, str] | None = None,
    ):
        """Initialize an AWS lambda function.

        :param name: function name
        :param description: a description of the function
        :param role: role to be asssumed during lambda execution
        :param code_bucket: bucket in which code for the function is found
        :param code_key: key in the previous bucket where the code is stored
        :param code_zipfile: inline code. it is needed when lambda code depends
            on stack informations only known at deployement.
        :param handler: handler name (i.e: entry point)
        :param code_version: code version
        :param timeout: maximum execution time (default: 3s)
        :param runtime: runtime to use
        :param memory_size: the amount of memory available to the function at
            runtime. The value can be any multiple of 1 MB.
        :param ephemeral_storage_size: The size of the function’s /tmp directory
            in MB. The default value is 512, but can be any whole number between
            512 and 10240 MB
        :param logs_retention_in_days: The number of days to retain the log events
            in the lambda log group
        :param reserved_concurrent_executions: The number of concurrent executions
            that are reserved for this function
        :param environment: Environment variables that are accessible from function
            code during execution
        """
        self.name = name
        self.description = description
        self.code_bucket = code_bucket
        self.code_key = code_key
        self.code_zipfile = code_zipfile
        self.code_version = code_version
        self.timeout = timeout
        self.runtime = runtime
        self.role = role
        self.handler = handler
        self.memory_size = memory_size
        self.ephemeral_storage_size = ephemeral_storage_size
        self.logs_retention_in_days = logs_retention_in_days
        self.reserved_concurrent_executions = reserved_concurrent_executions
        self.environment = environment

    def cfn_policy_document(self, stack: Stack) -> PolicyDocument:
        statements = [
            PolicyStatement(
                action=[
                    # Needed by CloudFormation to handle the function lifecycle
                    "lambda:CreateFunction",
                    "lambda:GetFunction",
                    "lambda:DeleteFunction",
                    "lambda:UpdateFunctionCode",
                    "lambda:UpdateFunctionConfiguration",
                    # Needed by resources referencing the function
                    "lambda:GetFunctionConfiguration",
                ],
                effect="Allow",
                resource=f"arn:aws:lambda:::function:{self.name}*",
            )
        ]
        if isinstance(self.role, GetAtt):
            logger.warning(f"cannot compute needed iam:PassRole for lambda {self.name}")
        else:
            if isinstance(self.role, Role):
                role_arn = f"arn:aws:iam::%(account)s:policy/{self.role.name}"
            else:
                role_arn = self.role

            # Allow user to pass role to the lambda
            statements.append(
                PolicyStatement(
                    action=["iam:PassRole"], effect="Allow", resource=role_arn
                )
            )
        return PolicyDocument(statements=statements)

    @property
    def arn(self) -> GetAtt:
        """Arn of the lambda funtion."""
        return GetAtt(name_to_id(self.name), "Arn")

    @property
    def ref(self) -> Ref:
        return Ref(name_to_id(self.name))

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return list of AWSObject associated with the construct."""
        return self.lambda_resources(
            code_bucket=self.code_bucket, code_key=self.code_key
        )

    def lambda_resources(
        self,
        code_bucket: str | None = None,
        code_key: str | None = None,
        image_uri: str | None = None,
    ) -> list[AWSObject]:
        """Return resource associated with the construct.

        :param code_bucket: bucket in which the lambda code is located
        :param code_key: location of the code in the bucket
        :param image_uri: URI of a container image in the Amazon ECR registry
        """
        # If code_bucket and code_key not provided use zipfile if
        # provided.

        params: dict[str, Any] = {}
        if code_bucket is not None and code_key is not None:
            code_params = {"S3Bucket": code_bucket, "S3Key": code_key}
            if self.code_version is not None:
                code_params["S3ObjectVersion"] = str(self.code_version)
        elif self.code_zipfile is not None:
            code_params = {"ZipFile": self.code_zipfile}
        elif image_uri:
            code_params = {"ImageUri": image_uri}
            params["PackageType"] = "Image"

        if isinstance(self.role, Role):
            role = self.role.arn
        else:
            role = self.role

        params.update(
            {
                "Code": awslambda.Code(**code_params),
                "Timeout": self.timeout,
                "Description": self.description,
                "Role": role,
                "FunctionName": self.name,
            }
        )

        if self.runtime is not None:
            params["Runtime"] = self.runtime

        if self.handler is not None:
            params["Handler"] = self.handler

        if self.memory_size is not None:
            params["MemorySize"] = self.memory_size

        if self.ephemeral_storage_size is not None:
            params["EphemeralStorage"] = awslambda.EphemeralStorage(
                Size=self.ephemeral_storage_size
            )
        if self.environment is not None:
            params["Environment"] = awslambda.Environment(Variables=self.environment)

        if self.reserved_concurrent_executions is not None:
            params["ReservedConcurrentExecutions"] = self.reserved_concurrent_executions

        result = [awslambda.Function(name_to_id(self.name), **params)]
        # If retention duration is given provide a log group.
        # If not provided the lambda creates a log group with
        # infinite retention.
        if self.logs_retention_in_days is not None:
            log_group = logs.LogGroup(
                name_to_id(f"{self.name}LogGroup"),
                DeletionPolicy="Retain",
                LogGroupName=f"/aws/lambda/{self.name}",
                RetentionInDays=self.logs_retention_in_days,
            )
            result.append(log_group)

        return result

    @staticmethod
    def lambda_log_group(lambda_name: str) -> Sub:
        """Return logroup arn for a given lambda.

        :param lambda_name: the lambda name
        """
        return Sub(
            "arn:aws:logs:${AWS::Region}:${AWS::AccountId}:log-group:/aws/lambda/"
            + lambda_name
        )

    @staticmethod
    def lambda_log_streams(lambda_name: str) -> Sub:
        """Return arn that matches all logstreams for a lambda log group.

        :param lambda_name: the lambda name
        """
        return Sub(
            "arn:aws:logs:${AWS::Region}:${AWS::AccountId}:log-group:/aws/lambda/"
            + lambda_name
            + ":*"
        )

    def invoke_permission(
        self,
        name_suffix: str,
        service: str,
        source_arn: str,
        source_account: str | None = None,
    ) -> awslambda.Permission:
        """Create a Lambda Permission object for a given service.

        :param name_suffix: a suffix used in the object name
        :param service: service name (without amazonaws.com domain name)
        :param source_arn: arn of the resource that can access the lambda
        :param source_account: account that holds the resource. This is
            mandatory only when using S3 as a service as a bucket arn is
            not linked to an account.
        :return: an AWSObject
        """
        params = {
            "Action": "lambda:InvokeFunction",
            "FunctionName": self.ref,
            "Principal": f"{service}.amazonaws.com",
            "SourceArn": source_arn,
        }
        if service == "s3":
            assert source_account is not None
        if source_account is not None:
            params["SourceAccount"] = source_account

        return awslambda.Permission(name_to_id(self.name + name_suffix), **params)


class DockerFunction(Function):
    """Lambda using a Docker image."""

    def __init__(
        self,
        name: str,
        description: str,
        role: str | GetAtt | Role,
        source_dir: str,
        repository_name: str,
        image_tag: str,
        timeout: int = 3,
        memory_size: int | None = None,
    ):
        """Initialize an AWS lambda function using a Docker image.

        :param name: function name
        :param description: a description of the function
        :param role: role to be asssumed during lambda execution
        :param source_dir: directory containing Dockerfile and dependencies
        :param repository_name: ECR repository name
        :param image_tag: docker image version
        :param timeout: maximum execution time (default: 3s)
        :param memory_size: the amount of memory available to the function at
            runtime. The value can be any multiple of 1 MB.
        """
        super().__init__(
            name=name,
            description=description,
            role=role,
            timeout=timeout,
            memory_size=memory_size,
        )
        self.source_dir: str = source_dir
        self.repository_name: str = repository_name
        timestamp = datetime.utcnow().strftime("%Y-%m-%d-%H-%M-%S-%f")
        self.image_tag: str = f"{image_tag}-{timestamp}"
        self.image_uri: str | None = None

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Compute AWS resources for the construct.

        Build and push the Docker image to ECR repository.
        Only push ECR image if stack is to be deployed.
        """
        if stack.dry_run:
            self.image_uri = "<dry_run_image_uri>"
        else:
            assert stack.deploy_session is not None
            self.image_uri = build_and_push_image(
                self.source_dir,
                self.repository_name,
                self.image_tag,
                stack.deploy_session,
            )

        return self.lambda_resources(image_uri=self.image_uri)


class PyFunction(Function):
    """Lambda with a Python runtime."""

    def __init__(
        self,
        name: str,
        description: str,
        role: str | GetAtt | Role,
        code_dir: str,
        handler: str,
        runtime: str,
        requirement_file: str | None = None,
        code_version: int | None = None,
        timeout: int = 3,
        memory_size: int | None = None,
        ephemeral_storage_size: int | None = None,
        logs_retention_in_days: int | None = 731,
        reserved_concurrent_executions: int | None = None,
        environment: dict[str, str] | None = None,
    ):
        """Initialize an AWS lambda function with a Python runtime.

        :param name: function name
        :param description: a description of the function
        :param role: role to be asssumed during lambda execution
        :param code_dir: directory containing the python code
        :param handler: name of the function to be invoked on lambda execution
        :param runtime: lambda runtime. It must be a Python runtime.
        :param requirement_file: requirement file for the application code.
            Required packages are automatically fetched (works only from linux)
            and packaged along with the lambda code
        :param code_version: code version
        :param timeout: maximum execution time (default: 3s)
        :param memory_size: the amount of memory available to the function at
            runtime. The value can be any multiple of 1 MB.
        :param ephemeral_storage_size: The size of the function’s /tmp directory
            in MB. The default value is 512, but can be any whole number between
            512 and 10240 MB
        :param logs_retention_in_days: The number of days to retain the log events
            in the lambda log group
        :param reserved_concurrent_executions: The number of concurrent executions
            that are reserved for this function
        :param environment: Environment variables that are accessible from function
            code during execution
        """
        assert runtime.startswith("python"), "PyFunction only accept Python runtimes"
        super().__init__(
            name=name,
            description=description,
            code_bucket=None,
            code_key=None,
            role=role,
            handler=handler,
            code_version=code_version,
            timeout=timeout,
            runtime=runtime,
            memory_size=memory_size,
            ephemeral_storage_size=ephemeral_storage_size,
            logs_retention_in_days=logs_retention_in_days,
            reserved_concurrent_executions=reserved_concurrent_executions,
            environment=environment,
        )
        self.code_dir = code_dir
        self.requirement_file = requirement_file

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Compute AWS resources for the construct."""
        assert isinstance(stack.s3_bucket, str)
        return self.lambda_resources(
            code_bucket=stack.s3_bucket,
            code_key=f"{stack.s3_key}{self.name}_lambda.zip",
        )

    def populate_package_dir(self, package_dir: str) -> None:
        """Copy user code into lambda package directory.

        :param package_dir: directory in which the package content is put
        """
        # Add lambda code
        sync_tree(self.code_dir, package_dir, delete=False)

    def create_data_dir(self, root_dir: str) -> None:
        """Create data to be pushed to bucket used by cloudformation for resources."""
        # Create directory specific to that lambda
        package_dir = os.path.join(root_dir, name_to_id(self.name), "package")

        # Install the requirements
        if self.requirement_file is not None:
            p = Run(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    f"--target={package_dir}",
                    "-r",
                    self.requirement_file,
                ],
                output=None,
            )
            assert p.status == 0

        # Copy user code
        self.populate_package_dir(package_dir=package_dir)

        # Create an archive
        create_archive(
            f"{self.name}_lambda.zip",
            from_dir=package_dir,
            dest=root_dir,
            no_root_dir=True,
        )

        # Remove temporary directory
        rm(package_dir, recursive=True)


class Py38Function(PyFunction):
    """Lambda using the Python 3.8 runtime."""

    def __init__(
        self,
        name: str,
        description: str,
        role: str | GetAtt | Role,
        code_dir: str,
        handler: str,
        requirement_file: str | None = None,
        code_version: int | None = None,
        timeout: int = 3,
        memory_size: int | None = None,
        ephemeral_storage_size: int | None = None,
        logs_retention_in_days: int | None = None,
        reserved_concurrent_executions: int | None = None,
    ):
        """Initialize an AWS lambda function using Python 3.8 runtime.

        See PyFunction for params description.
        """
        super().__init__(
            name=name,
            description=description,
            role=role,
            code_dir=code_dir,
            handler=handler,
            requirement_file=requirement_file,
            code_version=code_version,
            timeout=timeout,
            runtime="python3.8",
            memory_size=memory_size,
            ephemeral_storage_size=ephemeral_storage_size,
            logs_retention_in_days=logs_retention_in_days,
            reserved_concurrent_executions=reserved_concurrent_executions,
        )


class Alias(Construct):
    """A lambda alias."""

    def __init__(
        self,
        name: str,
        description: str,
        lambda_arn: str | GetAtt | Ref,
        lambda_version: str,
        provisioned_concurrency_config: awslambda.ProvisionedConcurrencyConfiguration
        | None = None,
        routing_config: awslambda.AliasRoutingConfiguration | None = None,
    ):
        """Initialize an AWS lambda alias.

        :param name: function name
        :param description: a description of the function
        :param lambda_arn: the name of the Lambda function
        :param lambda_version: the function version that the alias invokes
        :param provisioned_concurrency_config: specifies a provisioned
            concurrency configuration for a function's alias
        :param routing_config: the routing configuration of the alias
        """
        self.name = name
        self.description = description
        self.lambda_arn = lambda_arn
        self.lambda_version = lambda_version
        self.provisioned_concurrency_config = provisioned_concurrency_config
        self.routing_config = routing_config

    @property
    def ref(self) -> Ref:
        return Ref(name_to_id(self.name))

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return list of AWSObject associated with the construct."""
        params = {
            "Name": self.name,
            "Description": self.description,
            "FunctionName": self.lambda_arn,
            "FunctionVersion": self.lambda_version,
        }

        if self.provisioned_concurrency_config is not None:
            params["ProvisionedConcurrencyConfig"] = self.provisioned_concurrency_config

        if self.routing_config is not None:
            params["RoutingConfig"] = self.routing_config

        return [awslambda.Alias(name_to_id(self.name), **params)]


class Version(Construct):
    """A lambda version."""

    def __init__(
        self,
        name: str,
        description: str,
        lambda_arn: str | GetAtt | Ref,
        provisioned_concurrency_config: awslambda.ProvisionedConcurrencyConfiguration
        | None = None,
        code_sha256: str | None = None,
    ):
        """Initialize an AWS lambda version.

        :param name: version name
        :param description: a description for the version to override the description
            in the function configuration. Updates are not supported for this property
        :param lambda_arn: the name of the Lambda function
        :param provisioned_concurrency_config: specifies a provisioned concurrency
            configuration for a function's version. Updates are not supported for this
            property.
        :param code_sha256: only publish a version if the hash value matches the value
            that's specified. Use this option to avoid publishing a version if the
            function code has changed since you last updated it. Updates are not
            supported for this property
        """
        self.name = name
        self.description = description
        self.lambda_arn = lambda_arn
        self.provisioned_concurrency_config = provisioned_concurrency_config
        self.code_sha256 = code_sha256

    @property
    def ref(self) -> Ref:
        return Ref(name_to_id(self.name))

    @property
    def version(self) -> GetAtt:
        """Version of the lambda version."""
        return GetAtt(name_to_id(self.name), "Version")

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return list of AWSObject associated with the construct."""
        params = {
            "Description": self.description,
            "FunctionName": self.lambda_arn,
        }

        if self.provisioned_concurrency_config is not None:
            params["ProvisionedConcurrencyConfig"] = self.provisioned_concurrency_config

        if self.code_sha256 is not None:
            params["CodeSha256"] = self.code_sha256

        return [awslambda.Version(name_to_id(self.name), **params)]


class AutoVersion(Construct):
    """Automatic lambda versions."""

    def __init__(
        self,
        version: int,
        min_version: int | None = None,
        lambda_name: str | None = None,
        lambda_arn: str | GetAtt | Ref | None = None,
        lambda_function: Function | None = None,
        provisioned_concurrency_config: awslambda.ProvisionedConcurrencyConfiguration
        | None = None,
        code_sha256: str | None = None,
    ) -> None:
        """Create lambda versions from 1 to version included.

        When using this construct, you must provide either a lambda function or
        a lambda name plus arn.

        Parameters provisioned_concurrency_config and code_sha256 are only relevant
        for when creating a new version.

        :param version: number of the latest version
        :param min_version: minimum deployed version (default 1)
        :param lambda_name: the name of the Lambda function
        :param lambda_arn: the arn of the Lambda function
        :param lambda_function: the Lambda function
        :param provisioned_concurrency_config: specifies a provisioned concurrency
            configuration for a function's version. Updates are not supported for this
            property.
        :param code_sha256: only publish a version if the hash value matches the value
            that's specified. Use this option to avoid publishing a version if the
            function code has changed since you last updated it. Updates are not
            supported for this property
        """
        assert version > 0, "version should be greater than 0"
        assert (
            min_version is None or min_version > 0
        ), "min_version should be greater than 0"
        assert (
            min_version is None or min_version <= version
        ), "min_version can't be greater than version"
        assert lambda_function or (
            lambda_name is not None and lambda_arn is not None
        ), "either lambda_function or lambda_name plus lambda_arn should be provided"
        self.version = version
        self.min_version = min_version
        self.lambda_function = lambda_function
        self.lambda_name = lambda_function.name if lambda_function else lambda_name
        self.lambda_arn = lambda_function.arn if lambda_function else lambda_arn
        self.versions = [
            Version(
                name=f"{self.lambda_name}Version{i}",
                description=f"version {i} of {self.lambda_name} lambda",
                lambda_arn=self.lambda_arn,
            )
            for i in range(min_version if min_version is not None else 1, version + 1)
        ]
        self.latest.provisioned_concurrency_config = provisioned_concurrency_config
        self.latest.code_sha256 = code_sha256

    def get_version(self, number: int) -> Version | None:
        """Return a version.

        :param number: version number
        """
        return (
            self.versions[number - 1]
            if number > 0 and number <= len(self.versions)
            else None
        )

    @property
    def previous(self) -> Version:
        """Return the previous version.

        If there is only one version, then the latest version is
        returned.
        """
        # Last item if there is only 1 element
        # Last - 1 item if there is more than 1 elements
        return self.versions[-1] if len(self.versions) < 2 else self.versions[-2]

    @property
    def latest(self) -> Version:
        """Return the latest version."""
        return self.versions[-1]

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return list of AWSObject associated with the construct."""
        return self.versions


class BlueGreenAliasConfiguration(object):
    """Blue/Green alias configuration."""

    def __init__(
        self,
        version: str,
        name: str | None = None,
        provisioned_concurrency_config: awslambda.ProvisionedConcurrencyConfiguration
        | None = None,
        routing_config: awslambda.AliasRoutingConfiguration | None = None,
    ) -> None:
        """Configure a blue/green alias.

        :param version: lambda version pointed by the alias
        :param name: custom name for the alias
        :param provisioned_concurrency_config: specifies a provisioned
            concurrency configuration for a function's alias
        :param routing_config: the routing configuration of the alias
        """
        self.version = version
        self.name = name
        self.provisioned_concurrency_config = provisioned_concurrency_config
        self.routing_config = routing_config


class BlueGreenAliases(Construct):
    """Blue/Green aliases for a lambda."""

    def __init__(
        self,
        blue_config: BlueGreenAliasConfiguration,
        green_config: BlueGreenAliasConfiguration,
        lambda_name: str | None = None,
        lambda_arn: str | GetAtt | Ref | None = None,
        lambda_function: Function | None = None,
    ) -> None:
        """Create aliases for blue/green deployment of a lambda.

        Blue if the old version of the lambda used in production,
        while green is the new version used in development.

        :param blue_config: configuration for the blue alias
        :param green_config: configuration for the green alias
        :param lambda_name: the name of the Lambda function
        :param lambda_arn: the arn of the Lambda function
        :param lambda_function: the Lambda function
        """
        assert lambda_function or (
            lambda_name is not None and lambda_arn is not None
        ), "either lambda_function or lambda_name plus lambda_arn should be provided"
        self.blue_config = blue_config
        self.green_config = green_config
        self.lambda_function = lambda_function
        self.lambda_name = lambda_function.name if lambda_function else lambda_name
        self.lambda_arn = lambda_function.arn if lambda_function else lambda_arn

        def create_alias(
            config: BlueGreenAliasConfiguration, default_name: str
        ) -> Alias:
            """Return a new alias.

            :param config: alias configuration
            :param default_name: default alias name if none is specified
            """
            name = config.name if config.name is not None else default_name
            return Alias(
                name=name_to_id(f"{self.lambda_name}-{name}-alias"),
                description=f"{name} alias for {self.lambda_name} lambda",
                lambda_arn=self.lambda_arn,
                lambda_version=config.version,
                provisioned_concurrency_config=config.provisioned_concurrency_config,
                routing_config=config.routing_config,
            )

        self.aliases = [
            create_alias(config, default_name)
            for config, default_name in ((blue_config, "blue"), (green_config, "green"))
        ]

    @property
    def blue(self) -> Alias:
        """Return the blue alias."""
        return self.aliases[0]

    @property
    def green(self) -> Alias:
        """Return the green alias."""
        return self.aliases[1]

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return list of AWSObject associated with the construct."""
        return self.aliases



import pytest
from e3.aws.cfn.ec2 import VPC
from e3.aws.cfn.ec2.security import EgressRule, IngressRule, SecurityGroup


def test_security_group():
    vpc = VPC("vpc", cidr_block="10.10.0.0/16")
    rule1 = IngressRule("ssh", "10.10.1.1/32", description="ssh rule")
    rule2 = IngressRule("ip", "10.10.1.1/32", from_port=3389)
    rule2 = IngressRule("ip", "10.10.1.1/32", from_port=5000, to_port=5550)
    sg = SecurityGroup(
        "SecurityGroup", vpc, description="basic security group", rules=[rule1, rule2]
    )
    assert sg.properties

    sg.add_rule(EgressRule("ip", "10.10.1.1/32", from_port=80))
    assert sg.properties

    with pytest.raises(AssertionError):
        sg.add_rule("invalid object")



from __future__ import annotations
from e3.env import Env
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING
import re
import time
import uuid
import yaml
import logging

if TYPE_CHECKING:
    import botocore.client
    from typing import Iterator, Iterable, Callable, Any

VALID_STACK_NAME = re.compile("^[a-zA-Z][a-zA-Z0-9-]*$")
VALID_STACK_NAME_MAX_LEN = 128


def client(name: str) -> Callable:
    """Decorate a function to handle automatically AWS client retrieval.

    The function in input should take a mandatory argument called client.
    The function seen by the user will have an optional argument region
    to select the region in which the client is created.

    :param name: client name
    :type name: str
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            aws_env = Env().aws_env
            if "region" in kwargs:
                region = kwargs["region"]
                del kwargs["region"]
            else:
                region = aws_env.default_region
            client = aws_env.client(name, region=region)
            return func(*args, client=client, **kwargs)

        return wrapper

    return decorator


class AWSType(Enum):
    """Cloud Formation resource types."""

    EC2_EIP = "AWS::EC2::EIP"
    EC2_INSTANCE = "AWS::EC2::Instance"
    EC2_INTERNET_GATEWAY = "AWS::EC2::InternetGateway"
    EC2_LAUNCH_TEMPLATE = "AWS::EC2::LaunchTemplate"
    EC2_NAT_GATEWAY = "AWS::EC2::NatGateway"
    EC2_NETWORK_INTERFACE = "AWS::EC2::NetworkInterface"
    EC2_ROUTE = "AWS::EC2::Route"
    EC2_ROUTE_TABLE = "AWS::EC2::RouteTable"
    EC2_SECURITY_GROUP = "AWS::EC2::SecurityGroup"
    EC2_SUBNET = "AWS::EC2::Subnet"
    EC2_SUBNET_ROUTE_TABLE_ASSOCIATION = "AWS::EC2::SubnetRouteTableAssociation"
    EC2_VOLUME = "AWS::EC2::Volume"
    EC2_VPC = "AWS::EC2::VPC"
    EC2_VPC_ENDPOINT = "AWS::EC2::VPCEndpoint"
    EC2_VPC_GATEWAY_ATTACHMENT = "AWS::EC2::VPCGatewayAttachment"
    IAM_GROUP = "AWS::IAM::Group"
    IAM_ROLE = "AWS::IAM::Role"
    IAM_USER = "AWS::IAM::User"
    IAM_POLICY = "AWS::IAM::Policy"
    IAM_INSTANCE_PROFILE = "AWS::IAM::InstanceProfile"
    ROUTE53_HOSTED_ZONE = "AWS::Route53::HostedZone"
    ROUTE53_RECORDSET = "AWS::Route53::RecordSet"
    S3_BUCKET = "AWS::S3::Bucket"
    S3_BUCKET_POLICY = "AWS::S3::BucketPolicy"
    SERVICE_DISCOVERY_PRIVATE_DNS_NAMESPACE = (
        "AWS::ServiceDiscovery::PrivateDnsNamespace"
    )
    CODE_COMMIT_REPOSITORY = "AWS::CodeCommit::Repository"


class GetAtt(object):
    """Intrinsic function Fn::Getatt."""

    def __init__(self, name, attribute):
        """Initialize a Getatt instance.

        :param name: resource name
        :type name: str
        :param attribute: attribute name
        :type attribute: str
        """
        self.name = name
        self.attribute = attribute


class Ref(object):
    """Intrinsic function Fn::Ref."""

    def __init__(self, name):
        """Initialize a reference.

        :param name: resource name
        :type name: str
        """
        self.name = name


class Base64(object):
    """Intrinsic function Fn::Base64."""

    def __init__(self, content):
        """Initialize a base64 content.

        :param content: content to be encoded into base64
        :type content: str
        """
        self.content = content


class Join(object):
    """Intrinsic function Fn::Join."""

    def __init__(self, content, delimiter=""):
        """Initialize a Join object.

        :param content: a list
        :type content: list
        :param delimiter: a join delimiter
        :type delimiter: str
        """
        self.content = content
        self.delimiter = delimiter


class Sub(object):
    """Intrinsic function Fn::Sub."""

    def __init__(self, content, variables=None):
        """Initialize a Sub object.

        :param content: a string
        :type content: str
        :param variables: a dict asssociating a key to a value
        :type variables: dict(str, str)
        """
        self.content = content
        if variables:
            self.variables = dict(variables)
        else:
            self.variables = None


# Declare Yaml representer for intrinsic functions


def getatt_representer(dumper, data):
    return dumper.represent_scalar("!GetAtt", "%s.%s" % (data.name, data.attribute))


def ref_representer(dumper, data):
    return dumper.represent_scalar("!Ref", data.name)


def base64_representer(dumper, data):
    return dumper.represent_dict({"Fn::Base64": data.content})


def sub_representer(dumper, data):
    if data.variables:
        return dumper.represent_sequence("!Sub", [data.content, data.variables])
    else:
        return dumper.represent_scalar("!Sub", data.content)


def join_representer(dumper, data):
    return dumper.represent_sequence("!Join", [data.delimiter, data.content])


class CFNYamlDumper(yaml.Dumper):
    def __init__(self, *args, **kwargs):
        """Yaml dumper for cloud formation templates.

        See yaml.Dumper documentation.
        """
        super(CFNYamlDumper, self).__init__(*args, **kwargs)

        self.add_representer(GetAtt, getatt_representer)
        self.add_representer(Ref, ref_representer)
        self.add_representer(Base64, base64_representer)
        self.add_representer(Join, join_representer)
        self.add_representer(Sub, sub_representer)

    def ignore_aliases(self, data):
        """Ignore aliases."""
        return True


class Resource:
    """A CloudFormation resource."""

    # List of valid attribute names
    ATTRIBUTES: Iterable[str] = ()

    def __init__(self, name: str, kind: AWSType):
        """Initialize a resource.

        :param name: name of the resource (alphanumeric)
        :param kind: resource kind
        """
        assert isinstance(kind, AWSType), (
            "resource kind should be an AWSType: found %s" % kind
        )
        assert name.isalnum(), "resource name should be alphanumeric: found %s" % name
        self.name = name
        self.kind = kind
        self.depends = None

        # Track region in which the resource is created
        e = Env()
        if hasattr(e, "aws_env"):
            self.region = Env().aws_env.default_region
        else:
            self.region = "invalid-region"
        self.metadata: dict = {}

    def getatt(self, name):
        """Return an attribute reference.

        :param name: attribute name. should one of the valid attribute
            declared in ATTRIBUTES class variable
        :type name: str
        :return: a getatt object
        :rtype: e3.aws.cfn.types.GetAtt

        """
        assert name in self.ATTRIBUTES, "invalid attribute %s" % name
        return GetAtt(self.name, name)

    @property
    def ref(self):
        """Return a reference to the current resource.

        :return: a reference
        :rtype: e3.aws.cfn.types.Ref
        """
        return Ref(self.name)

    @property
    def properties(self):
        """Return the resource properties dict.

        :return: the resources Properties key value for the current resource.
        :rtype: dict
        """
        return {}

    def export(self) -> dict:
        """Export resource as a template fragment.

        :return: the dict representing the resources. The resulting dict can
            be serialized using Yaml to get a valid CloudFormation template
            fragment
        """
        result = {"Type": self.kind.value, "Properties": self.properties}
        if self.depends is not None:
            result["DependsOn"] = self.depends  # type: ignore
        if self.metadata:
            result["Metadata"] = self.metadata
        return result

    def create_data_dir(self, root_dir: str) -> None:
        """Put data in root_dir before export to S3 bucket referenced by the stack.

        :param root_dir: local directory in which data should be stored. Data will
            be then uploaded to an S3 bucket accessible from the template.
        """
        pass


class StackEventOperation(Enum):
    """Operations associated with stack events."""

    create = "CREATE"
    delete = "DELETE"
    update = "UPDATE"
    update_rollback = "UPDATE_ROLLBACK"
    import_resource = "IMPORT"
    import_rollback = "IMPORT_ROLLBACK"
    rollback = "ROLLBACK"

    def __str__(self) -> str:
        return {
            "CREATE": "creation",
            "DELETE": "deletion",
            "UPDATE": "update",
            "IMPORT": "import",
            "IMPORT_ROLLBACK": "import rollback",
            "UPDATE_ROLLBACK": "update rollback",
            "ROLLBACK": "rollback",
        }[self.value]


class StackEventState(Enum):
    """Operation states associated with stack events."""

    in_progress = "IN_PROGRESS"
    failed = "FAILED"
    complete = "COMPLETE"
    skipped = "SKIPPED"

    def __str__(self) -> str:
        return {
            "IN_PROGRESS": "started",
            "FAILED": "failed",
            "COMPLETE": "completed",
            "SKIPPED": "skipped",
        }[self.value]


class StackEventStatus:
    """Stack event status.

    This represents the combination of an operation and its current state

    :ivar operation: the operation name
    :ivar state: the operation state
    """

    def __init__(self, operation: StackEventOperation, state: StackEventState) -> None:
        """Initialize a stack event status.

        :param operation: an operation
        :param state: an operation state
        """
        self.operation = operation
        self.state = state

    @classmethod
    def from_str(cls, event_status_str: str) -> StackEventStatus:
        """Create a StackEventStatus based on string returned by AWS CFN.

        :param event_status_str: a string representing the status
        :return: a StackEventStatus
        """
        match = re.match(
            r"(CREATE|DELETE|UPDATE|IMPORT|IMPORT_ROLLBACK|UPDATE_ROLLBACK|ROLLBACK)_"
            r"(IN_PROGRESS|FAILED|COMPLETE|SKIPPED)",
            event_status_str,
        )
        assert match is not None, f"invalid event status {event_status_str}"
        return cls(
            operation=StackEventOperation(match.group(1)),
            state=StackEventState(match.group(2)),
        )

    def __str__(self) -> str:
        return f"{self.operation} {self.state}"


class StackEvent:
    """A stack event (see describe_stack_events API)."""

    def __init__(
        self,
        stack_id: str,
        event_id: str,
        stack_name: str,
        logical_resource_id: str,
        physical_resource_id: str,
        resource_type: str,
        timestamp: datetime,
        resource_status: StackEventStatus,
        client_token: str | None = None,
        resource_status_reason: str = "",
        resource_properties: str = "",
    ) -> None:
        """Create a stack event."""
        self.stack_id = stack_id
        self.event_id = event_id
        self.stack_name = stack_name
        self.logical_resource_id = logical_resource_id
        self.physical_resource_id = physical_resource_id
        self.resource_type = resource_type
        self.timestamp = timestamp
        self.resource_status = resource_status
        self.client_token = client_token
        self.resource_status_reason = resource_status_reason
        self.resource_properties = resource_properties

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> StackEvent:
        """Create a stack event from a dict as returned by AWS API."""
        return cls(
            stack_id=data["StackId"],
            event_id=data["EventId"],
            stack_name=data["StackName"],
            logical_resource_id=data["LogicalResourceId"],
            physical_resource_id=data["PhysicalResourceId"],
            resource_type=data["ResourceType"],
            timestamp=data["Timestamp"],
            resource_status=StackEventStatus.from_str(data["ResourceStatus"]),
            client_token=data.get("ClientRequestToken", None),
            resource_status_reason=data.get("ResourceStatusReason", ""),
            resource_properties=data.get("ResourceProperties", ""),
        )

    def __str__(self) -> str:
        return (
            f"{self.logical_resource_id:<32}: {self.resource_type:<32}: "
            + f"{str(self.resource_status):<16} ({self.resource_status_reason})"
        )


class Stack(object):
    """A CloudFormation stack."""

    def __init__(
        self,
        name: str,
        description: str | None = None,
        cfn_role_arn: str | None = None,
        s3_bucket: str | None = None,
        s3_key: str | None = None,
    ):
        """Initialize a stack.

        :param name: stack name
        :param description: a description of the stack
        :param cfn_role_arn: Arn of the role to be assumed by cloudformation. If
            None then use user role. In the future role_arn should be mandatory in
            order to avoid giving users too much rights.
        :param s3_bucket: s3 bucket used to store data needed by the stack
        :param s3_key: s3 prefix in s3_bucket in which data is stored
        """
        assert (
            re.match(VALID_STACK_NAME, name) and len(name) <= VALID_STACK_NAME_MAX_LEN
        ), ("invalid stack name: %s" % name)
        self.resources: dict[str, Resource | Stack] = {}
        self.name = name

        # In most cfn calls name and id can be used for StackName parameter. On first
        # call to state the real stack_id will be stored in self.stack_id ensuring
        # that the stack can be still refered to even when operation like delete are
        # used.
        self.stack_id = name
        self.description = description
        self.s3_bucket = s3_bucket
        self.s3_key = s3_key

        # Emit a warning to the user if no role is passed for Cloud Formation
        if cfn_role_arn is None:
            logging.warning(
                "Consider using a separate role for CloudFormation to "
                "reduce permissions needed by the entity in charge of "
                "deploying the stack (see Stack cfn_role_arn parameter)"
            )
        self.cfn_role_arn = cfn_role_arn
        self.creation_date = datetime.now().timestamp()

        # The uuid is used by create and create_change_set. It allows then
        # to track for example events associated with that deployment
        self.uuid = str(uuid.uuid1(clock_seq=int(1000 * time.time())))
        self.latest_read_event: StackEvent | None = None

    def add(self, element: Stack | Resource) -> Stack:
        """Add a resource or merge a stack.

        :param element: if a resource add the resource to the stack. If a stack
            merge its resources into the current stack.
        :return: the current stack
        """
        assert isinstance(element, Resource) or isinstance(element, Stack), (
            "a resource or a stack is expected. got %s" % element
        )
        assert element.name not in self.resources, (
            "resource already exist: %s" % element.name
        )
        self.resources[element.name] = element
        return self

    def __iadd__(self, element):
        """Add a resource or merge a stack.

        :param element: if a resource add the resource to the stack. If a stack
            merge its resources into the current stack.
        :type element: Stack | Resources
        :return: the current stack
        :rtype: Stack
        """
        return self.add(element)

    def __getitem__(self, key):
        if key not in self.resources:
            raise KeyError
        return self.resources[key]

    def __contains__(self, key):
        return key in self.resources

    def create_data_dir(self, root_dir: str) -> None:
        """Populate directory that will be exported into a S3 bucket for the stack.

        :param root_dir: temporary local directory
        """
        for resource in self.resources.values():
            resource.create_data_dir(root_dir)

    def export(self) -> dict:
        """Export stack as dict.

        :return: a dict that can be serialized as YAML to produce a template
        :rtype: dict
        """
        resources = {}
        for resource in self.resources.values():
            if isinstance(resource, Resource):
                assert resource.name not in resources
                resources[resource.name] = resource.export()
            else:
                # resource is a stack
                stack_resources = resource.export()["Resources"]
                for k, v in stack_resources.items():
                    assert k not in resources
                    resources[k] = v

        result = {"AWSTemplateFormatVersion": "2010-09-09", "Resources": resources}
        if self.description is not None:
            result["Description"] = self.description
        return result

    @property
    def body(self) -> str:
        """Export stack as a CloudFormation template.

        :return: a valid CloudFormation template
        :rtype: str
        """
        return yaml.dump(self.export(), Dumper=CFNYamlDumper)

    @client("cloudformation")
    def create(
        self,
        client: botocore.client.Client,
        url: str | None = None,
        wait: bool = False,
    ) -> None:
        """Create a stack.

        :param client: a botocore client
        param url: url of the template body in S3. When not None this suppose
            the user has uploaded the template body on S3 first at the given
            url. Use S3 to refer to the template body rather than using inline
            version allows to use template of size up to 500Ko instead of
            50Ko.
        :param wait: if True wait for creation completion
        """
        parameters = {
            "StackName": self.name,
            "Capabilities": ["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM"],
        }

        if url is None:
            parameters["TemplateBody"] = self.body
        else:
            parameters["TemplateURL"] = url

        if self.cfn_role_arn is not None:
            parameters["RoleARN"] = self.cfn_role_arn

        parameters["ClientRequestToken"] = self.uuid
        client.create_stack(**parameters)

        if wait:
            logging.info("Waiting for stack creation...")
            logging.info(f"Done (status: {self.wait()})")

    @client("cloudformation")
    def wait(self, client: botocore.client.Client) -> str:
        status = self.state()
        while "PROGRESS" in status["StackStatus"]:
            for event in self.events(mark_as_read=True):
                logging.info(str(event))
            time.sleep(5.0)
            status = self.state()

        # Get last eents
        for event in self.events(mark_as_read=True):
            logging.info(str(event))
        return status["StackStatus"]

    def exists(self) -> bool:
        """Check if a given stack exists.

        :return: True if it does, False otherwise
        """
        try:
            self.state()
            return True
        except Exception:
            # Documentation does not specify the right exception that is raised
            # by botocore.
            return False

    @client("cloudformation")
    def state(self, client):
        """Return state of the stack on AWS."""
        result = client.describe_stacks(StackName=self.stack_id)["Stacks"][0]
        if self.stack_id != result["StackId"]:
            self.stack_id = result["StackId"]
        return result

    @client("cloudformation")
    def validate(self, client, url=None):
        """Validate a template.

        :param client: a botocore client
        :type client: botocore.client.Client. This parameter is handled by the
            decorator
        param url: url of the template body in S3. When not None this suppose
            the user has uploaded the template body on S3 first at the given
            url. Use S3 to refer to the template body rather than using inline
            version allows to use template of size up to 500Ko instead of
            50Ko.
        :type url: str | None
        """
        if url is None:
            return client.validate_template(TemplateBody=self.body)
        else:
            return client.validate_template(TemplateURL=url)

    @client("cloudformation")
    def create_change_set(self, name, client, url=None):
        """Create a change set.

        This creates a difference between the state of the stack on AWS servers
        and the current one generated with e3-aws.

        :param client: a botocore client. This parameter is handled by the
            decorator
        :type client: botocore.client.Client
        :param name: name of the changeset
        :type name: str
        :param url: url of the template body in S3. When not None this suppose
            the user has uploaded the template body on S3 first at the given
            url. Use S3 to refer to the template body rather than using inline
            version allows to use template of size up to 500Ko instead of
            50Ko.
        :type url: str | None
        """
        parameters = {
            "ChangeSetName": name,
            "StackName": self.name,
            "Capabilities": ["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM"],
        }

        if self.cfn_role_arn is not None:
            parameters["RoleARN"] = self.cfn_role_arn

        if url is None:
            return client.create_change_set(
                TemplateBody=self.body,
                **parameters,
            )
        else:
            return client.create_change_set(
                TemplateURL=url,
                **parameters,
            )

    @client("cloudformation")
    def describe_change_set(self, name, client):
        """Describe a change set.

        Retrieve status of a given changeset

        :param client: a botocore client
        :type client: botocore.client.Client
        :param name: name of the changeset
        :type name: str
        """
        return client.describe_change_set(ChangeSetName=name, StackName=self.name)

    @client("cloudformation")
    def delete_change_set(self, name, client):
        """Delete a change set.

        :param client: a botocore client
        :type client: botocore.client.Client
        :param name: name of the changeset
        :type name: str
        """
        return client.delete_change_set(ChangeSetName=name, StackName=self.name)

    @client("cloudformation")
    def delete(self, client: botocore.client.Client, wait: bool = False) -> None:
        """Delete a stack.

        Delete a stack. Note that operation is aynchron

        :param client: a botocore client
        :type client: botocore.client.Client
        :param wait: if True wait for complete deletion
        """
        # Ensure to fill stack_id
        try:
            self.state()
        except Exception:
            logging.error(f"Stack {self.name} does not exist")
            return

        client.delete_stack(StackName=self.name, ClientRequestToken=self.uuid)
        if wait:
            logging.info("Wait for stack deletion")
            logging.info(f"Done (status: {self.wait()})")

    @client("cloudformation")
    def events(
        self,
        client: botocore.client.BaseClient,
        failed_only: bool = False,
        mark_as_read: bool = True,
    ) -> Iterator[StackEvent]:
        """Return non read events.

        :param failed_only: return only failed events
        :param mark_as_read: if True, all events read won't be returned on next
            calls to events method.
        """
        from e3.aws import iterate

        latest_read_event = self.latest_read_event
        if mark_as_read:
            self.latest_read_event = None

        for element in iterate(
            client.describe_stack_events, key="StackEvents", StackName=self.stack_id
        ):
            event = StackEvent.from_dict(element)

            # Update marker for last event read
            if mark_as_read and self.latest_read_event is None:
                self.latest_read_event = event

            if event.client_token != self.uuid:
                # Event is not associated with current operations. Skip
                continue
            elif (
                latest_read_event is not None
                and latest_read_event.timestamp == event.timestamp
            ):
                # Event is already read so stop iterating.
                break
            elif failed_only and event.resource_status != StackEventState.failed:
                continue
            else:
                yield event

    @client("cloudformation")
    def cost(self, client):
        """Compute cost of the stack (estimation).

        :param client: a botocore client
        :type client: botocore.client.BaseClient
        """
        return client.estimate_template_cost(TemplateBody=self.body)

    @client("cloudformation")
    def execute_change_set(self, client, changeset_name, wait=False):
        """Execute a changeset.

        :param client: a botocore client
        :type client: botocore.client.BaseClient
        :param changeset_name: name of the changeset to apply
        :type changeset_name: str
        :param wait: whether to wait for the completion of the command
        :type wait: bool
        """
        client.execute_change_set(
            ChangeSetName=changeset_name,
            StackName=self.name,
            ClientRequestToken=self.uuid,
        )

        if wait:
            logging.info("Waiting for stack update...")
            logging.info(f"Done (status: {self.wait()})")

    @client("cloudformation")
    def resource_status(self, client, in_progress_only=True):
        """Return status of each resources of the stack.

        The state of the stack taken is the one pushed on AWS (after a call
        to create for example).

        :param client: a botocore client
        :type client: botocore.client.BaseClient
        :param in_progress_only: if True return only resources that are in
            one of the "PROGRESS" state (deletion, creation, ...)
        :type in_progress_only: bool
        :return: a dict associating a resource logical name to a status name
        :rtype: dict
        """
        aws_result = client.describe_stack_resources(StackName=self.name)
        assert "StackResources" in aws_result
        result = {}
        for res in aws_result["StackResources"]:
            if "PROGRESS" in res["ResourceStatus"] or not in_progress_only:
                result[res["LogicalResourceId"]] = res["ResourceStatus"]
        return result

    @client("cloudformation")
    def enable_termination_protection(self, client):
        """Enable termination protection for a stack."""
        aws_result = self.state()
        if aws_result["EnableTerminationProtection"]:
            print("Stack termination protection is already enabled")
            return

        aws_result = client.update_termination_protection(
            EnableTerminationProtection=True, StackName=self.name
        )
        assert "StackId" in aws_result
        print("Stack termination protection enabled")

    @client("cloudformation")
    def set_stack_policy(self, stack_policy_body, client):
        """Set a stack policy.

        :param stack_policy_body: stack policy body to apply
        :type stack_policy_body: str
        """
        aws_result = client.get_stack_policy(StackName=self.name)

        if stack_policy_body != aws_result["StackPolicyBody"]:
            print("Stack policy has been modified")

            client.set_stack_policy(
                StackName=self.name, StackPolicyBody=stack_policy_body
            )
        else:
            print("Stack policy already up-to-date")



from e3.aws.cfn.iam import Allow, PolicyDocument
from e3.aws.cfn.s3 import AccessControl, Bucket, BucketPolicy


def test_create_bucket_policy():
    p = PolicyDocument().append(
        (Allow(to="s3:GetObject", on=["arn:aws:s3:::mybucket"]))
    )
    b = BucketPolicy(name="mypolicy", bucket="mybucket", policy_document=p)
    assert b.properties
    b = BucketPolicy(name="mypolicy", bucket=Bucket(name="mybucket"), policy_document=p)
    assert b.properties


def test_create_bucket():
    b = Bucket(
        "mybucket",
        access_control=AccessControl.PRIVATE,
        bucket_name="myname",
        versioning=True,
    )
    assert b.properties



from __future__ import annotations
from functools import cached_property
from typing import TYPE_CHECKING

from troposphere import ec2, GetAtt, Ref, Tags

from e3.aws import name_to_id
from e3.aws.troposphere import Construct
from e3.aws.troposphere.iam.policy_document import PolicyDocument

if TYPE_CHECKING:
    from troposphere import AWSObject
    from e3.aws.troposphere import Stack


class InternetGateway(Construct):
    """InternetGateway construct.

    Provide an internet gateway attached to a given VPC and a route table routing
    traffic from given subnets to the gateway.
    """

    def __init__(
        self,
        name_prefix: str,
        vpc: ec2.vpc,
        subnets: list[ec2.subnet] | None = None,
        route_table: ec2.RouteTable | None = None,
    ):
        """Initialize InternetGateway construct.

        :param name_prefix: prefix for cloudformation resource names
        :param vpc: VPC to attach to InternetGateway
        :param subnets: subnets from wich traffic should be routed to the internet
            gateway. Only needed if a route table is not provided. If provided a
            route table is created and associated to these subnets.
        :param route_table: Add route to internet gateway to this route. If no
            route is provided, one is created.
        """
        self.vpc = vpc
        self.subnets = subnets
        self.name_prefix = name_prefix
        self._route_table = route_table
        self.add_route_table_to_stack = route_table is None

    @property
    def route_table(self) -> ec2.RouteTable:
        """Return route table to which the route to IGW is added."""
        if self._route_table is None:
            self._route_table = ec2.RouteTable(
                name_to_id(f"{self.name_prefix}-igw-route-table"), VpcId=Ref(self.vpc)
            )
        return self._route_table

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return resources associated with the construct."""
        igw = ec2.InternetGateway(name_to_id(f"{self.name_prefix}-igw"))
        attachement = ec2.VPCGatewayAttachment(
            name_to_id(f"{self.name_prefix}-igw-attachement"),
            InternetGatewayId=Ref(igw),
            VpcId=Ref(self.vpc),
        )
        route = ec2.Route(
            name_to_id(f"{self.name_prefix}-igw-route"),
            RouteTableId=Ref(self.route_table),
            DestinationCidrBlock="0.0.0.0/0",
            GatewayId=Ref(igw),
        )
        result = [igw, attachement, route]

        # If a new route table has to be created associate it with provided subnets
        if self.add_route_table_to_stack:
            result.append(self.route_table)
            assert self.subnets is not None
            result.extend(
                [
                    ec2.SubnetRouteTableAssociation(
                        name_to_id(f"{self.name_prefix}-{num}"),
                        RouteTableId=Ref(self.route_table),
                        SubnetId=Ref(subnet),
                    )
                    for subnet, num in zip(  # noqa: B905
                        self.subnets, range(len(self.subnets))
                    )
                ]
            )

        return result


class VPCEndpointsSubnet(Construct):
    """VPCEndpointsSubnet Construct.

    Provide a subnet with Interface VPC endpoints and a security group
    configured to authorize access to endpoints from the VPC.
    """

    def __init__(
        self,
        name: str,
        region: str,
        cidr_block: str,
        vpc: ec2.vpc,
        interface_endpoints: list[tuple[str, PolicyDocument | None]] | None = None,
    ) -> None:
        """Initialize VPCEndpointsSubnet Construct.

        :param name: name of the subnet
        :param region: AWS region where to deploy the Subnet
        :param vpc_endpoint_cidr_block: The IPv4 CIDR block assigned to the subnet
        :param vpc: attach the subnet to this vpc
        :param interface_endpoint: list of (<service_name>, <endpoint_policy_document>)
            tuples for each interface endpoint to create in the vpc endpoints subnet.
        """
        self.name = name
        self.region = region
        self.cidr_block = cidr_block
        self.vpc = vpc
        self.has_ses_endpoint = False

        if interface_endpoints:
            self.interface_endpoints = interface_endpoints
            if "email-smtp" in [endpoint[0] for endpoint in self.interface_endpoints]:
                self.has_ses_endpoint = True
        else:
            self.interface_endpoints = []

    @cached_property
    def subnet(self) -> ec2.Subnet:
        """Return a Subnet for VPC endpoints."""
        subnet_name = f"{self.name}Subnet"
        return ec2.Subnet(
            name_to_id(subnet_name),
            VpcId=Ref(self.vpc),
            CidrBlock=self.cidr_block,
            Tags=Tags({"Name": subnet_name}),
        )

    @cached_property
    def security_group(self) -> ec2.SecurityGroup:
        """Return a security group for VPC endpoints."""
        return ec2.SecurityGroup(
            name_to_id(f"{self.name}SecurityGroup"),
            GroupDescription=f"{self.name} vpc endpoints security group",
            SecurityGroupEgress=[],
            SecurityGroupIngress=[],
            VpcId=Ref(self.vpc),
        )

    @cached_property
    def ses_security_group(self) -> ec2.SecurityGroup:
        """Return a security group for SES VPC endpoint."""
        return ec2.SecurityGroup(
            name_to_id(f"{self.name}SESSecurityGroup"),
            GroupDescription=f"{self.name} SES vpc endpoint security group",
            SecurityGroupEgress=[
                ec2.SecurityGroupRule(
                    CidrIp=self.vpc.CidrBlock,
                    IpProtocol="-1",
                )
            ],
            SecurityGroupIngress=[
                ec2.SecurityGroupRule(
                    CidrIp=self.vpc.CidrBlock,
                    FromPort="587",
                    ToPort="587",
                    IpProtocol="tcp",
                )
            ],
            VpcId=Ref(self.vpc),
        )

    @cached_property
    def https_ingress_rule(self) -> ec2.SecurityGroupIngress:
        """Return Ingress rule allowing HTTPS traffic from the VPC."""
        return ec2.SecurityGroupIngress(
            name_to_id(f"{self.name}IngressFromVPC"),
            CidrIp=self.vpc.CidrBlock,
            FromPort="443",
            ToPort="443",
            IpProtocol="tcp",
            GroupId=Ref(self.security_group),
        )

    @cached_property
    def https_egress_rule(self) -> ec2.SecurityGroupEgress:
        """Return Egress rule allowing HTTPS traffic to the VPC."""
        return ec2.SecurityGroupEgress(
            name_to_id(f"{self.name}EgressToVPC"),
            CidrIp=self.vpc.CidrBlock,
            FromPort="443",
            ToPort="443",
            IpProtocol="tcp",
            GroupId=Ref(self.security_group),
        )

    @cached_property
    def default_egress_rule(self) -> ec2.SecurityGroupEgress:
        """Return egress that disables default egress Rule."""
        return ec2.SecurityGroupEgress(
            name_to_id(f"{self.name}DefaultEgress"),
            CidrIp=self.cidr_block,
            IpProtocol="-1",
            GroupId=Ref(self.security_group),
        )

    @cached_property
    def interface_vpc_endpoints(self) -> list[ec2.VPCEndpoint]:
        """Return interface endpoints."""
        endpoints = []

        for service_name, pd in self.interface_endpoints:
            if pd is not None:
                opt_params = {"PolicyDocument": pd.as_dict}
            else:
                opt_params = {}

            if service_name == "email-smtp":
                security_group_id = Ref(self.ses_security_group)
            else:
                security_group_id = Ref(self.security_group)

            endpoints.append(
                ec2.VPCEndpoint(
                    name_to_id(f"{service_name}Endpoint"),
                    PrivateDnsEnabled="true",
                    SecurityGroupIds=[security_group_id],
                    ServiceName=f"com.amazonaws.{self.region}.{service_name}",
                    SubnetIds=[Ref(self.subnet)],
                    VpcEndpointType="Interface",
                    VpcId=Ref(self.vpc),
                    **opt_params,
                )
            )
        return endpoints

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Construct and return VPCEndpointsSubnet resources."""
        result = [
            self.subnet,
            self.security_group,
            self.default_egress_rule,
            self.https_egress_rule,
            self.https_ingress_rule,
            *self.interface_vpc_endpoints,
        ]

        if self.has_ses_endpoint:
            result.append(self.ses_security_group)

        return result


class Subnet(Construct):
    """Subnet Construct.

    Add a NAT gateway if needed if the subnet is public (i.e a route_table to an
    internet gateway must be provided). For a private subnet a route to a NAT
    gateway can be created.
    """

    def __init__(
        self,
        name: str,
        vpc: ec2.VPC,
        cidr_block: str,
        availability_zone: str,
        internet_gateway: InternetGateway | None = None,
        use_nat: bool = False,
        nat_to: ec2.NatGateway | None = None,
    ):
        """Initialize Subnet construct.

        :param name: name of the Subnet
        :param vpc: vpc in which to create the subnet
        :param cidr_block: address range of the subnet. Should be a subnet of the
            vpc address range (no check done).
        :param availability_zone: the availability zone of the subnet
        :param internet_gateway: Internet Gateway to route traffic to. It has to
            be provided for a public subnet. if provided the route table associated
            with the internet gateway will be associated to the subnet.
        :param use_nat: if True then add a NAT gateway that can be used by
            private subnets. To do so the subnet must be public i.e internet_gateway
            must be provided.
        :param nat_to: if is_public is False and nat_to is not None, then create
            a route to the NAT gateway.
        """
        self.name = name
        self.cidr_block = cidr_block
        self.vpc = vpc
        self.availability_zone = availability_zone
        self.internet_gateway = internet_gateway
        self.use_nat = use_nat
        self.nat_to = nat_to

        if use_nat:
            assert (
                internet_gateway is not None
            ), "a NAT Gateway can only be added to a public subnet"

        self._subnet: ec2.Subnet | None = None
        self._route_table: ec2.RouteTable | None = None
        self._nat_gateway: ec2.NatGateway | None = None
        self._nat_eip: ec2.EIP | None = None

    @cached_property
    def subnet(self) -> ec2.Subnet:
        """Return a private subnet."""
        return ec2.Subnet(
            name_to_id(self.name),
            VpcId=Ref(self.vpc),
            CidrBlock=self.cidr_block,
            Tags=Tags({"Name": self.name}),
            AvailabilityZone=self.availability_zone,
        )

    @cached_property
    def route_table(self) -> ec2.RouteTable:
        """Return a route table for this subnet."""
        if self.internet_gateway:
            # By default only one route table is used to route traffic
            # from public subnets to the Internet Gateway.
            return self.internet_gateway.route_table
        else:
            return ec2.RouteTable(
                name_to_id(f"{self.name}RouteTable"), VpcId=Ref(self.vpc)
            )

    @cached_property
    def route_table_assoc(self) -> ec2.SubnetRouteTableAssociation:
        """Return association of route table to this subnet."""
        return ec2.SubnetRouteTableAssociation(
            name_to_id(f"{self.name}RouteTableAssoc"),
            RouteTableId=Ref(self.route_table),
            SubnetId=Ref(self.subnet),
        )

    @cached_property
    def nat_eip(self) -> ec2.EIP | None:
        """Return an elastic IP for the NAT gateway."""
        return ec2.EIP(name_to_id(f"{self.name}EIP"))

    @cached_property
    def nat_gateway(self) -> ec2.NatGateway | None:
        """Return a NAT gateway for this subnet."""
        return ec2.NatGateway(
            name_to_id(f"{self.name}NAT"),
            AllocationId=GetAtt(self.nat_eip, "AllocationId"),
            SubnetId=Ref(self.subnet),
        )

    @cached_property
    def ID(self) -> Ref:
        """Return subnet's ID."""
        return Ref(self.subnet)

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return resources associated with the Subnet construct."""
        result = [self.subnet, self.route_table_assoc]

        if not self.internet_gateway:
            result.append(self.route_table)
        if self.use_nat:
            result.extend([self.nat_gateway, self.nat_eip])

        if self.nat_to:
            result.append(
                ec2.Route(
                    name_to_id(f"{self.name}NATRoute"),
                    RouteTableId=Ref(self.route_table),
                    DestinationCidrBlock="0.0.0.0/0",
                    NatGatewayId=Ref(self.nat_to),
                )
            )

        return result


class VPC(Construct):
    """VPC Construct.

    Provide a VPC with:
    * a primary and an optional secondary public subnets
    * an optional private subnet which can have a route to the NAT Gateway of
      the primary public subnets
    * a endpoints subnet with VPC endpoints configured according to arguments
    * an optional s3 endpoint for the private subnet or the primary public subnet
      if there is no private subnet or NAT configured
    """

    def __init__(
        self,
        name: str,
        region: str = "eu-west-1",
        cidr_block: str = "10.10.0.0/16",
        private_subnet_cidr_block: str | None = "10.10.0.0/18",
        private_subnet_az: str = "eu-west-1a",
        public_subnet_cidr_block: str = "10.10.64.0/18",
        public_subnet_az: str = "eu-west-1a",
        secondary_public_subnet_cidr_block: str | None = "10.10.128.0/18",
        secondary_public_subnet_az: str = "eu-west-1b",
        vpc_endpoints_subnet_cidr_block: str = "10.10.192.0/18",
        nat_gateway: bool = False,
        s3_endpoint_policy_document: PolicyDocument | None = None,
        interface_endpoints: list[tuple[str, PolicyDocument | None]] | None = None,
        tags: dict[str, str] | None = None,
    ) -> None:
        """Initialize VPC Construct.

        :param name: name of the VPC
        :param region: region where to deploy the VPC
        :param cidr_block: the primary IPv4 CIDR block for the VPC
        :param private_subnet_cidr_block: The IPv4 CIDR block assigned to the
            private subnet. If None no private subnet is created.
        :param public_subnet_cidr_block: The IPv4 CIDR block assigned to the
            public subnet
        :param secondary_public_subnet_cidr_block: The IPv4 CIDR block for an
            optional secondary public subnet
        :param vpc_endpoint_cidr_block: The IPv4 CIDR block assigned to the VPC
            endpoints subnet
        :param nat_gateway: set it to True to add a NatGateway.
        :param s3_endpoint_policy_document: policy for s3 endpoint. If none is
            given no s3 endpoint is created.
        :param interface_endpoint: list of (<service_name>, <endpoint_policy_document>)
            tuples for each interface endpoint to create in the vpc endpoints subnet.
        :param tags: tags for the VPC
        """
        self.name = name
        self.region = region
        self.cidr_block = cidr_block
        self.nat_gateway = nat_gateway
        self.vpc_endpoints_subnet_cidr_block = vpc_endpoints_subnet_cidr_block
        self.s3_endpoint_policy_document = s3_endpoint_policy_document

        self.tags: dict[str, str]
        if tags is not None:
            self.tags = tags
        else:
            self.tags = {}

        # Add a VPC and associate an InternetGateway to it.
        # Traffic is routed from public subnets to the InternetGateway
        # thought a route table that is associated to all public subnets
        self.vpc = ec2.VPC(
            name_to_id(self.name),
            CidrBlock=self.cidr_block,
            EnableDnsHostnames="true",
            EnableDnsSupport="true",
            Tags=Tags({"Name": self.name, **self.tags}),
        )
        self.public_subnets_route_table = ec2.RouteTable(
            name_to_id(f"{self.name}PublicSubnetsRouteTable"), VpcId=Ref(self.vpc)
        )
        self.internet_gateway = InternetGateway(
            name_prefix=self.name,
            vpc=self.vpc,
            route_table=self.public_subnets_route_table,
        )

        # Add public subnets
        self.public_subnet = Subnet(
            name=f"{self.name}PublicSubnet",
            vpc=self.vpc,
            cidr_block=public_subnet_cidr_block,
            internet_gateway=self.internet_gateway,
            use_nat=True,
            availability_zone=public_subnet_az,
        )
        self.secondary_public_subnet: Subnet | None = None
        if secondary_public_subnet_cidr_block:
            self.secondary_public_subnet = Subnet(
                name=f"{self.name}SecondaryPublicSubnet",
                vpc=self.vpc,
                cidr_block=secondary_public_subnet_cidr_block,
                internet_gateway=self.internet_gateway,
                availability_zone=secondary_public_subnet_az,
            )

        # Add a private subnet if requested and route outgoing traffic to the
        # primary public subnet NAT Gateway
        self.private_subnet: Subnet | None = None
        if private_subnet_cidr_block:
            self.private_subnet = Subnet(
                name=f"{self.name}PrivateSubnet",
                vpc=self.vpc,
                cidr_block=private_subnet_cidr_block,
                nat_to=self.public_subnet.nat_gateway,
                availability_zone=private_subnet_az,
            )
        else:
            self.private_subnet = None
        self._security_group: ec2.SecurityGroup | None = None
        self.vpc_endpoints_subnet = VPCEndpointsSubnet(
            name=f"{self.name}VPCEndpointsSubnet",
            region=region,
            cidr_block=vpc_endpoints_subnet_cidr_block,
            vpc=self.vpc,
            interface_endpoints=interface_endpoints,
        )

    @cached_property
    def main_subnet(self) -> ec2.Subnet:
        """Return the subnet where instances/task that access Internet should run.

        If there is no NAT gateway, instances/tasks should be run in the public
        subnet to have an Internet Access.
        """
        if self.private_subnet and self.nat_gateway:
            return self.private_subnet.subnet
        else:
            return self.public_subnet.subnet

    # Security groups and traffic control
    @cached_property
    def security_group(self) -> ec2.SecurityGroup:
        """Return main security group."""
        sg_name = f"{self.name}SecurityGroup"
        return ec2.SecurityGroup(
            name_to_id(f"{self.name}SecurityGroup"),
            GroupDescription=f"{self.name} main security group",
            SecurityGroupEgress=[],
            SecurityGroupIngress=[],
            VpcId=Ref(self.vpc),
            Tags=Tags({"Name": sg_name}),
        )

    @cached_property
    def egress_to_vpc_endpoints(self) -> list[ec2.SecurityGroupRule]:
        """Return egress rules allowing traffic to VPC endpoints.

        This is an helper function to create security groups with permissions to
        access VPC endpoints.
        """
        rules = [
            ec2.SecurityGroupRule(
                DestinationSecurityGroupId=Ref(
                    self.vpc_endpoints_subnet.security_group
                ),
                Description="Allows traffic to the subnet holding VPC "
                "interface endpoints",
                FromPort="443",
                ToPort="443",
                IpProtocol="tcp",
            ),
        ]
        if self.s3_endpoint_policy_document:
            rules.append(
                ec2.SecurityGroupRule(
                    Description="Allows traffic to S3 VPC endpoint",
                    DestinationPrefixListId="pl-6da54004",
                    FromPort="443",
                    ToPort="443",
                    IpProtocol="tcp",
                )
            )
        if self.vpc_endpoints_subnet.has_ses_endpoint:
            rules.append(
                ec2.SecurityGroupRule(
                    DestinationSecurityGroupId=Ref(
                        self.vpc_endpoints_subnet.ses_security_group
                    ),
                    Description="Allows traffic to the subnet holding the SES VPC "
                    "interface endpoints",
                    FromPort="587",
                    ToPort="587",
                    IpProtocol="tcp",
                )
            )
        return rules

    @cached_property
    def endpoints_egress_rule(self) -> ec2.SecurityGroupEgress:
        """Return egress allowing traffic to VPC interface endpoints ."""
        return ec2.SecurityGroupEgress(
            name_to_id(f"{self.name}EndpointsEgress"),
            DestinationSecurityGroupId=Ref(self.vpc_endpoints_subnet.security_group),
            Description="Allows traffic to the subnet holding VPC interface endpoints",
            FromPort="443",
            ToPort="443",
            IpProtocol="tcp",
            GroupId=Ref(self.security_group),
        )

    @cached_property
    def s3_egress_rule(self) -> ec2.SecurityGroupEgress | None:
        """Return security group egress rule allowing outgoing S3 traffic."""
        if self.s3_endpoint_policy_document:
            return ec2.SecurityGroupEgress(
                name_to_id(f"{self.name}S3Egress"),
                Description="Allows traffic though S3 VPC endpoint",
                DestinationPrefixListId="pl-6da54004",
                FromPort="443",
                ToPort="443",
                IpProtocol="tcp",
                GroupId=Ref(self.security_group),
            )
        else:
            return None

    @cached_property
    def s3_route_table(self) -> ec2.RouteTable:
        """Return the route table for the s3 endpoint.

        Plug it to the route_table of the subnet where instances/tasks should be running
        """
        if self.nat_gateway and self.private_subnet:
            return self.private_subnet.route_table
        else:
            return self.public_subnet.route_table

    @cached_property
    def s3_vpc_endpoint(self) -> ec2.VPCEndPoint | None:
        """Return S3 VPC Endpoint.

        Note that this endpoint is also needed when using ECR as ECR stores
        images on S3.
        """
        if self.s3_endpoint_policy_document:
            return ec2.VPCEndpoint(
                name_to_id(f"{self.name}S3Endpoint"),
                PolicyDocument=self.s3_endpoint_policy_document.as_dict,
                RouteTableIds=[Ref(self.s3_route_table)],
                ServiceName=f"com.amazonaws.{self.region}.s3",
                VpcEndpointType="Gateway",
                VpcId=Ref(self.vpc),
            )
        else:
            return None

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return VPC Construct resources."""
        return [
            el
            for el in (
                self.vpc,
                self.security_group,
                self.private_subnet,
                self.public_subnet,
                self.public_subnets_route_table,
                self.secondary_public_subnet,
                self.internet_gateway,
                self.vpc_endpoints_subnet,
                self.endpoints_egress_rule,
                self.s3_vpc_endpoint,
                self.s3_egress_rule,
            )
            if el is not None
        ]



from __future__ import absolute_import, division, print_function

import pytest
from botocore.stub import ANY, Stubber
from e3.aws import AWSEnv, default_region
from e3.aws.cfn import Stack
from e3.aws.cfn.s3 import Bucket


def test_stack_create():
    s = Stack(name="teststack")
    s.body
    assert True

    with pytest.raises(AssertionError):
        # Create a stack with an invalid name
        s = Stack(name="test_stack")


def test_stack_compose():
    s = Stack(name="teststack")
    s2 = Stack(name="teststack2")
    s2.add(Bucket("bucket1")).add(Bucket("bucket2"))
    s += s2
    assert len(s.export()["Resources"]) == 2


def test_create_stack():
    s = Stack(name="teststack")

    aws_env = AWSEnv(regions=["us-east-1"])
    with default_region("us-east-1"):
        cfn_client = aws_env.client("cloudformation", region="us-east-1")

        stubber = Stubber(cfn_client)
        stubber.add_response(
            "create_stack",
            {},
            {
                "Capabilities": ["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM"],
                "StackName": "teststack",
                "ClientRequestToken": ANY,
                "TemplateBody": ANY,
            },
        )
        stubber.add_response(
            "create_stack",
            {},
            {
                "Capabilities": ["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM"],
                "StackName": "teststack",
                "ClientRequestToken": ANY,
                "TemplateURL": ANY,
            },
        )
        with stubber:
            s.create()
            s.create(url="noprotocol://nothing")


def test_create_change_set():
    s = Stack(name="teststack")

    aws_env = AWSEnv(regions=["us-east-1"])
    with default_region("us-east-1"):
        cfn_client = aws_env.client("cloudformation", region="us-east-1")

        stubber = Stubber(cfn_client)
        stubber.add_response(
            "create_change_set",
            {},
            {
                "Capabilities": ["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM"],
                "StackName": "teststack",
                "ChangeSetName": "name1",
                "TemplateBody": ANY,
            },
        )
        stubber.add_response(
            "create_change_set",
            {},
            {
                "Capabilities": ["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM"],
                "ChangeSetName": "name2",
                "StackName": "teststack",
                "TemplateURL": ANY,
            },
        )
        with stubber:
            s.create_change_set("name1")
            s.create_change_set("name2", url="noprotocol://nothing")


def test_create_change_set_role_arn():
    s = Stack(name="teststack", cfn_role_arn="arn:aws:iam::123456789012:role/S3Access")

    aws_env = AWSEnv(regions=["us-east-1"])
    with default_region("us-east-1"):
        cfn_client = aws_env.client("cloudformation", region="us-east-1")

        stubber = Stubber(cfn_client)
        stubber.add_response(
            "create_change_set",
            {},
            {
                "Capabilities": ["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM"],
                "StackName": "teststack",
                "ChangeSetName": "name1",
                "TemplateBody": ANY,
                "RoleARN": "arn:aws:iam::123456789012:role/S3Access",
            },
        )
        with stubber:
            s.create_change_set("name1")


def test_validate():
    s = Stack(name="teststack")

    aws_env = AWSEnv(regions=["us-east-1"])
    with default_region("us-east-1"):
        cfn_client = aws_env.client("cloudformation", region="us-east-1")

        stubber = Stubber(cfn_client)
        stubber.add_response("validate_template", {}, {"TemplateBody": ANY})
        stubber.add_response("validate_template", {}, {"TemplateURL": ANY})
        with stubber:
            s.validate()
            s.validate(url="noprotocol://nothing")



import boto3
import time  # noqa: F401


def handler(event, context):
    path = []
    client = boto3.client("cloudfront")  # noqa: F841
    for items in event["Records"]:
        if items["s3"]["object"]["key"] == "index.html":
            path.append("/")
        else:
            path.append("/" + items["s3"]["object"]["key"])



"""Provide AWS Config configuration rules."""
from __future__ import annotations
from dataclasses import dataclass, field
from typing import TYPE_CHECKING


from troposphere import AWSObject, config

from e3.aws import name_to_id
from e3.aws.troposphere import Construct


if TYPE_CHECKING:
    from typing import Any
    from e3.aws.troposphere import Stack


@dataclass(frozen=True)
class ConfigRule(Construct):
    """Define a configuration rule.

    :param name: name of the rule
    :param source_identifier: predefined identifier of an existing AWS Config
        managed rule
    :param description: rule description
    :param input_parameters: input parameters passed to the AWS Config rule
        Lambda function
    :param scope: defines which resources can trigger an evaluation for the rule
    """

    name: str
    source_identifier: str = ""
    description: str = ""
    input_parameters: dict[str, Any] = field(default_factory=dict)
    scope: dict[str, Any] = field(default_factory=dict)

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return troposphere objects defining the configuration rule."""
        return [
            config.ConfigRule(
                name_to_id(self.name),
                ConfigRuleName=self.name,
                Description=self.description,
                InputParameters=self.input_parameters,
                Scope=config.Scope.from_dict(
                    name_to_id(f"configscope{self.name}"), self.scope
                ),
                Source=config.Source.from_dict(
                    name_to_id(f"configsource{self.name}"),
                    {"Owner": "AWS", "SourceIdentifier": self.source_identifier},
                ),
                DependsOn="ConfigRecorder",
            )
        ]


@dataclass(frozen=True)
class S3ConfigRule(ConfigRule):
    """Define a S3 Configuration rule.

    :param name: name of the rule
    :param source_identifier: predefined identifier of an existing AWS Config
        managed rule
    :param description: rule description
    :param input_parameters: input parameters passed to the AWS Config rule
        Lambda function
    """

    name: str
    source_identifier: str
    description: str
    input_parameters: dict = field(default_factory=dict)

    scope: dict = field(
        default_factory=lambda: {"ComplianceResourceTypes": ["AWS::S3::Bucket"]},
        init=False,
    )


S3BucketPublicWriteProhibited = S3ConfigRule(
    name="s3-bucket-public-write-prohibited",
    source_identifier="S3_BUCKET_PUBLIC_WRITE_PROHIBITED",
    description=(
        "Checks that your S3 buckets do not allow public write access."
        "If an S3 bucket policy or bucket ACL allows public write access, "
        "the bucket is noncompliant."
    ),
)

S3BucketPublicReadProhibited = S3ConfigRule(
    name="s3-bucket-public-read-prohibited",
    source_identifier="S3_BUCKET_PUBLIC_READ_PROHIBITED",
    description=(
        "Checks that your S3 buckets do not allow public read access."
        "If an S3 bucket policy or bucket ACL allows public read access, "
        "the bucket is noncompliant."
    ),
)

S3BucketServerSideEncryptionEnabled = S3ConfigRule(
    name="s3-bucket-server-side-encryption-enabled",
    source_identifier="S3_BUCKET_SERVER_SIDE_ENCRYPTION_ENABLED",
    description=(
        "Checks that your Amazon S3 bucket either has S3 default encryption "
        "enabled or that the S3 bucket policy explicitly denies put-object "
        "requests without server side encryption."
    ),
)


S3BucketSSLRequestsOnly = S3ConfigRule(
    name="s3-bucket-ssl-requests-only",
    source_identifier="S3_BUCKET_SSL_REQUESTS_ONLY",
    description=(
        "Checks whether S3 buckets have policies that require requests to "
        "use Secure Socket Layer (SSL)."
    ),
)

IAMUserNoPoliciesCheck = ConfigRule(
    name="iam-user-no-policies-check",
    source_identifier="IAM_USER_NO_POLICIES_CHECK",
    description=(
        "Checks that none of your IAM users have policies attached. "
        "IAM users must inherit permissions from IAM groups or roles."
    ),
    scope={"ComplianceResourceTypes": ["AWS::IAM::User"]},
)



from ipaddress import AddressValueError, IPv4Network, IPv6Network
import logging
import requests
from e3.aws.cfn.ec2.security import Ipv4EgressRule, SecurityGroup

# This is the static address at which AWS publish the list of ip-ranges
# used by its services.
IP_RANGES_URL = "https://ip-ranges.amazonaws.com/ip-ranges.json"


def amazon_security_groups(name, vpc):
    """Create a dict of security group authorizing access to aws services.

    As the number of rules per security group is limited to 60,
    we create blocks of 60 rules.

    :param vpc: vpc in which to create the group
    :type vpc: VPC
    :return: a dict of security groups indexed by name
    :rtype: dict(str, SecurityGroup)
    """

    def select_region(ip_range_record):
        """Select the VPN region and the us-east-1 region.

        Note that some global interface (e.g. sts) are only available in
        the us-east-1 region.
        """
        return ip_range_record["region"] in (vpc.region, "us-east-1")

    ip_ranges = requests.get(IP_RANGES_URL).json()["prefixes"]

    # Retrieve first the complete list of ipv4 ip ranges for a given region
    amazon_ip_ranges = {
        k["ip_prefix"]
        for k in ip_ranges
        if select_region(k) and "ip_prefix" in k and k["service"] == "AMAZON"
    }

    # Substract the list of ip ranges corresponding to services
    # that we do no need to access to or that we access through VPC
    # endpoints to limit the number of security groups and rules.
    services_used = ("AMAZON", "S3")
    removable_ip_ranges = {
        k["ip_prefix"]
        for k in ip_ranges
        if select_region(k) and "ip_prefix" in k and k["service"] not in services_used
    }
    amazon_ip_ranges -= removable_ip_ranges

    # Authorize https on the resulting list of ip ranges
    sgs = {}
    i = 0
    limit = 60
    sg_name = name + str(i)
    sg = SecurityGroup(sg_name, vpc, description="Allow access to amazon services")
    sgs[sg_name] = sg
    for ip_range in amazon_ip_ranges:
        if len(sg.egress + sg.ingress) == limit:
            i += 1
            sg_name = name + str(i)
            sg = SecurityGroup(
                sg_name, vpc, description="Allow acces to amazon services"
            )
            sgs[sg_name] = sg
        sg.add_rule(Ipv4EgressRule("https", ip_range))
    return sgs


def github_security_groups(name, vpc, protocol):
    """Create a dict of security group authorizing access to github services.

    As the number of rules per security group is limited to 50,
    we create blocks of 50 rules.

    :param vpc: vpc in which to create the group
    :type vpc: VPC
    :param protocol: protocol to allow (https, ssh)
    :type protocol: str
    :return: a dict of security groups indexed by name
    :rtype: dict(str, SecurityGroup)
    """
    ip_ranges = requests.get("https://api.github.com/meta").json()["git"]

    # Authorize ssh on the resulting list of ip ranges
    sgs = {}
    i = 0
    limit = 50
    sg_name = name + str(i)
    sg = SecurityGroup(sg_name, vpc, description="Allow access to github")
    sgs[sg_name] = sg
    for ip_range in ip_ranges:
        try:
            IPv4Network(ip_range)
        except AddressValueError:
            IPv6Network(ip_range)
            logging.info(f"Skipping IPv6 range {ip_range} for github access SG")
            continue

        if len(sg.egress + sg.ingress) == limit:
            i += 1
            sg_name = name + str(i)
            sg = SecurityGroup(
                sg_name, vpc, description=f"Allow access to GitHub {protocol}"
            )
            sgs[sg_name] = sg
        sg.add_rule(Ipv4EgressRule(protocol, ip_range))
    return sgs
